# The Broad Institute
# SOFTWARE COPYRIGHT NOTICE AGREEMENT
# This software and its documentation are copyright 2003-2012 by the
# Broad Institute/Massachusetts Institute of Technology.
# All rights are reserved.
#
# This software is supplied without any warranty or guaranteed support
# whatsoever. Neither the Broad Institute nor MIT can be responsible for
# its use, misuse, or functionality.

ATARiS <- function(...) {
  sink(file=stdout(), type = "message")
  args <- list(...)
  tmp <- getParams(args)
  params <- tmp$params
  libdir <- tmp$libdir
  
  with (params, {
    lnames <- names(params)
    log.file <- paste(params$results.dir, params$results.prefix, "_log.txt", sep="")
    write(c("Parameters for shRNA analysis ran on:", date()), file = log.file)
    for (i in 1:length(lnames)) {
      write("----------", append = T, file = log.file)
      write(paste(lnames[i], ":", sep=""), append = T, file = log.file)
      write("          ", append = T, file = log.file)
      write(params[[i]], append = T, file = log.file)
    }
    write("-----------------------------------------------------", append  = T, file = log.file)
  }) # with params

  # set random number generator seed
  set.seed(params$random.seed)

  # Load input gct file
  loadFiles.out <- RIGER.loadFiles(params)
  m <- loadFiles.out$merged.array
  rownames(m) <-  loadFiles.out$constructs.set
  colnames(m) <- loadFiles.out$sample.names
  
  cons.gene.map <- loadFiles.out$cons.gene.map
  names(cons.gene.map) <- loadFiles.out$constructs.set

  cat(paste("Input file contains data for", ncol(m), "samples.",
            length(unique(cons.gene.map)), "genes targeted by",
            nrow(m), "RNAi reagents.\n\n"))

  
  ### populate the significance cutoffs data structure (from file or generate)
#  source(paste(libdir,'factorize.l1.R', sep=''))

  cat("Computing null distributions for gene solutions:\n")
  s.values <- generateSignificanceCutoffs(params, m, cons.gene.map)

  
  ### run the analysis
  print("Analyzing...")
  analysis.time <- proc.time()
  gene.solutions <- getGenesSolutions(m, cons.gene.map, s.values, gene.list=params$genes.to.analyze,
                                      num.cores=params$num.cores, fit.method=params$fit.method,
                                      sol.fit.score.method=params$sol.fit.score.method, params$hp.fit.score.method,
                                      params$positive.As, params$sol.pval.method, params$null.significance,
                                      sol.accept.method=params$sol.accept.method, min.A.value=params$min.A.value)
  
  results <- list(gene.solutions=gene.solutions, m=m,
                  num.samples=ncol(m), cons.gene.map=cons.gene.map, s.values=s.values, params=params)

  cat("\n")
#  cat("Time it took to analyze:\n")
#  print(proc.time() - analysis.time)

  if (!is.null(params$genes.to.analyze)) {
    restricted.cons.gene.map <- cons.gene.map[cons.gene.map %in% params$genes.to.analyze]
  } else {
    restricted.cons.gene.map <- cons.gene.map
  }

  cat("Organizing and saving results...\n")
  
  gene.table <- extractGeneTable(results)
  results <- c(results, gene.table=list(gene.table))
  out.file <- paste(params$results.dir, params$identifier, '.gene.table.txt', sep="")
  write.table(gene.table, file=out.file, sep='\t', quote=F)

  total.num.solutions <- sum(gene.table[,"num.solutions"])
  Gs <- extractGTable(results, total.num.solutions)
  results <- c(results, Gs=list(Gs), total.num.solutions=total.num.solutions)
#  out.file <- paste(params$results.dir, params$identifier, '.Gs.Rdata', sep="")
#  save(Gs, file=out.file)

  cat("Extracting solution table...")
  solutions <- extractSolutionTable(results)
  results <- c(results, solutions=list(solutions))
  cat("Done.\n")

  out.file <- paste(params$results.dir, params$identifier, '.Gs.gct', sep="")
  write.gct(Gs, descs=results$solutions[,"gene.name"], filename=out.file)

  writeSummary(results, params$results.dir, params$identifier)

  cat("Extracting shRNA table...\n")
  hpTable <- extractHairpinTable(results)

    ### compute c-scores
  cat("Computing c-scores for all genes...\n")
#  source(paste(libdir,'c-scores.R', sep=''))
  cscores.results <- compCScoreAllGenes(m,
                                        restricted.cons.gene.map,
                                        summarization.method=params$cscores.summarization.method,
                                        corr.method=params$cscores.corr.method,
                                        num.null.values=params$cscores.num.null.values,
                                        num.cores=params$num.cores, hp.table=hpTable)
  cscores <- cscores.results$cscores
  gene.cor.pvals <- cscores.results$gene.cor.pvals
  results <- c(results, cscores=list(cscores), gene.cor.pvals=list(gene.cor.pvals))
  
  hpTable <- cbind(hpTable, results$cscores[rownames(hpTable), 1:3]) # merge with cscores table
  results <- c(results, hp.table=list(hpTable))
  out.file <- paste(params$results.dir, params$identifier, '.hp.table.txt', sep="")
  write.table(hpTable, file=out.file, sep='\t', quote=F, row.names=F, col.names=T)
  
#  print("Saving results R file...")
#  save(results, params, file=paste(params$results.dir, params$identifier, '.results.Rdata', sep=""))

}


getParams <- function(args) {

  params <- DefaultParams()
  for(i in 1:length(args[[1]])) {
    arg <- args[[1]][i]
    flag <- substring(arg, 0, 2)
    value <- substring(arg, 3, nchar(arg))
    #print(flag)
    #print(value)
    if(flag=='-i') {
      params$identifier <- value
    } else if(flag=='-f') {
      params$shRNA.files <- value
    } else if(flag=='-s') {
      params$random.seed <- as.numeric(value)
    } else if(flag=='-l') {
      libdir <- value
    } else if (flag=='-n') {
      params$null.significance <- as.numeric(value)
    } else if (flag=='-m') {
      params$min.A.value <- as.numeric(value)
    } else {
      stop("ERROR: input option not recognized!")
    }
  }

  params$results.prefix <- params$identifier          

  return(list(params=params, libdir=libdir))
}


DefaultParams <- function() {
  params <- list()
  params <- within(params, {
    # user params
    shRNA.files <- NULL # gct file to load
    identifier <-  NULL # to be determined by user
    random.seed <- 12345
    exclude.genes <- c("LUCIFERASE", "lacZ", "RFP", "rfp", "GFP", "clonetechGfp", "promegaLuc","NA", "", "0", "pgw")
    exclude.cons <- c("empty", "PGW/empty", "PGW")

    # general params
    results.dir <-   "./" # always current dir for GP modules
    analysis.type <- "single.sample"
    num.cores <- 1

    fit.method <- "l1"
    sol.fit.score.method <- 'worst.hp.score'
    hp.fit.score.method <- 'spearman.cor'
    positive.As <- T
    min.A.value <- 0.3
    
    # load main data file, filter, standardize
    trunc.val <- +Inf              # Truncation value for survival scores
    file.format <- "GCT"
    norm.type <- "none"
    cntrl.plot <- F
    cntrl.genes <- NULL 
    cntrl.genes2 <- NULL

    genes.to.analyze <- NULL

    ### parameters concerning the cutoffs for solution fitness score 
    s.values.file <- NULL # use default

    # cscores computation parameters
    cscores.summarization.method <- "combined.pval"
    cscores.corr.method <- "spearman"
    cscores.num.null.values <- 10000
  
    sol.pval.method <- 'null.one.level'

    sol.accept.method <- "null.pval"
    null.significance <- 0.15
  
    num.null.points <- 1000
    min.num.null.points <- 200

}) # within params

  return(params)
}


# gets a matrix with each row a vector. returns a vector where the sum of pearson correlation coeffs to all the
# given vectors is maximal.
getMaxSumCorVector <- function(m) {

  # if m is only one hairpin
  if (is.null(dim(m)))
    return(m)
  
  # scales a vector to a unit vector
  scaleToUnit <- function(x) {  x / sqrt(sum(x^2, na.rm=T)) }
  
  m.mc <- t(apply(m, 1, function(x) x - mean(x, na.rm=T))) # the vectors, mean centered
  m.umc <-  t(apply(m.mc, 1, scaleToUnit)) # the vectors, scaled to unit circle

  g <- apply(m.umc, 2, sum, na.rm=T) # this is the vector we're looking for (before unit normalization)
  
  scaleToUnit(g) # g is also mean-centered as is sum of mean-centered vectors
}


# returns a data frame with two columns - corr and p.value,
# containing the Pearson corr score between x and each of the rows of m along with their p.value
getCorrSignificance <- function(x, m) {
  t(apply(m, 1, function(v) { tmp=cor.test(x, v, alternative='g'); c(tmp$estimate, p.value=tmp$p.value)} ))
}

# Returns a matrix with containing in each row indices of hps from base.mat to be used as the hps of one random gene.
# The matrix has 'num.genes' rows.
generateRandomGenes <- function(base.mat, num.genes, hps.per.gene) {
  total.hps <- num.genes * hps.per.gene
  rand.m <- matrix(sample(nrow(base.mat), total.hps, replace=T), nrow=num.genes, ncol=hps.per.gene)

  rand.m
}


# given a data matrix, generates 'num.genes' random genes of size 'hps.per.gene'. For each gene
# an lsfit solution is found. The R^2 value for the solution is returned for each random gene.
generateNullRsquare <- function(base.mat, num.null.points, hps.per.gene, fit.method) {
  num.genes <- num.null.points

  rand.genes <- generateRandomGenes(base.mat, num.genes, hps.per.gene)
  Rs <- numeric(num.genes)
  
  for (i in 1:num.genes) {
    hps <- rand.genes[i,]
    rand.gene <- base.mat[hps,] # the data of the random gene

    sol <- factanal.ccr(t(rand.gene), factors=1, method=fit.method, iter.max=50, drawplot=F, print=F, orth=F)
    Rs[i] <- getRsquare(t(rand.gene), sol, method=fit.method)
  }

  Rs
}


# given a data matrix, generates 'num.null.points' random genes of size 'hps.per.gene'. For each gene
# a consensus is computed and a fit score is computed. Then the hp with the least agreement (worst hp fit score)
# to the solution is removed and the procedure is repeated until we are left with min.sol.size hps to find a
# consensus to.
# In each iteration the solution score is saved. The solution score is determined by sol.fit.score.method.
# An array of size hps.per.gene x num.null.points is returned. Element [i,j] is the solution score for the i-th
# random gene when j hps (out of the original hps.per.gene hps) are used. (j goes from hps.per.gene to min.sol.size)
# min.sol.size - the minimum number of hps to use to generate a solution. used to be 2, but might be interested in
# exploring only null values for solutions of size 5, for example.
generateNullSolutionScores <- function(base.mat, num.null.points, hps.per.gene, fit.method,
                                       num.cores=1, sol.fit.score.method, hp.fit.score.method,
                                       positive.As, sol.pval.method, min.sol.size=2, null.significance,
                                       sol.accept.method, min.A.value) {
#  cat(paste("Generating null gene scores for", num.null.points, 'random genes using the method', fit.method, '...\n'))
  
  rand.genes <- generateRandomGenes(base.mat, num.null.points, hps.per.gene)
  

  # the following function is called below to generate gene scores for one random gene 
  inner <- function(rand.gene.ID) {
    hps <- rownames(base.mat)[rand.genes[rand.gene.ID, ]]
    rand.gene.scores <- rep(NA, length(hps))

    while (length(hps) >= min.sol.size) {
      num.hps <- length(hps)
      
      rand.gene <- base.mat[hps,]
      if (fit.method %in% c("lsfit", "l1fit", "wl1fit", "wl1", "l1")) {
        
        sol <- findGeneFit(rand.gene, s.values=NULL, rownames(rand.gene), fit.method, sol.fit.score.method,
                           hp.fit.score.method, positive.As, sol.pval.method=NULL,
                           null.significance=null.significance,
                           sol.accept.method=sol.accept.method,
                           min.A.value=min.A.value)

        rand.gene.scores[num.hps] <- sol$sol.fit.score
        hp.to.remove <- sol$next.hp.to.remove
      } else if (fit.method == "cor.w.G") {
        G <- getMaxSumCorVector(rand.gene)
        hps.agreement <- getHPsAgreement(rand.gene, G)
        
        rand.gene.scores[num.hps] <- min(hps.agreement)
        
        hp.to.remove <- hps[which.min(hps.agreement)]
      }
      
      hps <- setdiff(hps, hp.to.remove)
    }
    rand.gene.scores
  } # inner()

  # loop on all random genes
  if (num.cores > 1) {
    require(multicore, quietly=TRUE)
    all.random.genes.scores <- mclapply(1:num.null.points, inner, mc.cores=num.cores)
  } else {
    all.random.genes.scores <- lapply(1:num.null.points, inner)
  }

  solution.scores <- matrix(NA, nrow=num.null.points, ncol=hps.per.gene) # the gene scores of all null points
  # fit results in a matrix instead of a list
  for (i in 1:num.null.points) {
    solution.scores[i, ] <- all.random.genes.scores[[i]]
  }

  
  solution.scores
}



# For each given gene size, generate null distribution of gene fitness score using the appropriate fitting method.
# Returns for each gene size a vector with the 'significance' (or 1-'significance') quantile significance value
# for each number of included hps, from 2 to gene.size.
#
# base.mat - data matrix. hps in rows.
# num.null.points - size of the sample, i.e. how many min(cor.w.G) to compute. or: a vector of size max(gene.sizes) with
#   the i-th element being the number of null points to generate for genes of this size.
# gene.sizes - a vector of gene sizes for which to create null agreement scores.
getSignificanceCutoff <- function(base.mat, num.null.points, gene.sizes,
                                  null.significance, num.cores=1, fit.method,
                                  sol.fit.score.method, hp.fit.score.method, positive.As=NULL,
                                  sol.pval.method, sol.accept.method, min.A.value) {
  if (length(num.null.points) == 1) {
    num.null.points <- rep(num.null.points, max(gene.sizes))
  }
  
  s.values <- vector(mode="list", length=max(gene.sizes))

  for (gene.size in gene.sizes) {
    stopifnot(gene.size > 1)

    if (sol.pval.method == 'null.two.levels') {
      min.sol.size <- 2
    } else if (sol.pval.method == 'null.one.level') {
      min.sol.size <- gene.size
    } else {
      stop("unsupported sol.pval.method value")
    }
    
    cat(paste("Computing null", null.significance, "significance values for genes of size", gene.size, "...\n"))
    nullD <- generateNullSolutionScores(base.mat, num.null.points[gene.size], gene.size, fit.method,
                                        num.cores=num.cores, sol.fit.score.method, hp.fit.score.method,
                                        positive.As, sol.pval.method, min.sol.size=min.sol.size,
                                        null.significance=null.significance,
                                        sol.accept.method=sol.accept.method,
                                        min.A.value=min.A.value)

# compute quantiles just to display as informative output
    if (fit.method == "cor.w.G") {
      s.value <- apply(nullD, 2, quantile, probs=1-null.significance, na.rm=T) # s.value is a vector
    } else if (fit.method %in% c("lsfit", "l1fit", "wl1fit", "wl1", "l1")) {
      s.value <- apply(nullD, 2, quantile, probs=null.significance, na.rm=T)        
    } else {
      stop(paste('unsuppoerted method type', fit.method))
    }

#    cat(paste(s.value, "(n=", num.null.points[gene.size], ")\n"))

    s.values[[gene.size]] <- nullD 
  }

  s.values
}


# returns a values for each hairpin signifying the agreement of its profile with the given consensus profil.
# m - gene matrix. hp depletion values in rows. samples in columns.
# G - a consensus profile
getHPsAgreement <- function(m, G) {
  apply(m, 1, function(x) cor(x, G, use="na.or.complete", method="pearson"))

}




# given hp data and possible solution G, computes the best hp efficacies (A) and hp constant effects (beta)
# for each hp. (Best in Least Sum of Squares sense)
# m - hp data in rows (with hp ID as row.names)
# G - consensus profile
getBestHPparams <- function(m, G) {
  A <- beta <- vector(mode="numeric", length=nrow(m))
  names(A) <- names(beta) <- row.names(m)
  
  # lm() will fail if G and a hairpin don't have at least 2 data points in common.
  G.good.samples <- names(G)[!is.na(G)]
  ok.hps <- names(A)[apply(m, 1, function(x) sum(!is.na(x[G.good.samples])) >= 2)] # hps with at least three data points shared with G
  ex.hps <- setdiff(names(A), ok.hps)
  A[ex.hps] <- beta[ex.hps] <- NA
  
    # given m and the Gs, find the best beta.h and A to fit the data.
    # there are nh independent least-square fittings going on here, one for each hp.
    # I.e., for each hp its values in the m matrix are estimated using the Gs.
    # No hp weights here because for each hp we're looking for the minimum err.
   #  compute lm for each hp separately as na.omit eremoves all columns with any NAs....
  for (h in ok.hps) {
    coeffs <- lm(m[h,] ~ G, na.action='na.omit')$coefficients
    beta[h] <- coeffs[1]
    A[h] <- coeffs[2]
  }

  
#  if (length(ok.hps) >= 1) {
#    coeffs <- lm(t(m[ok.hps,]) ~ G, na.action='na.omit')$coefficients
#    beta[ok.hps] <- coeffs[1,]
#    A[ok.hps] <- coeffs[2,]
#  }
  
    list(A=A, beta=beta)
}



# Returns normalized hp efficacies (A) and consensus profile (G) so the max(A)=1 and A*G stays the same.
normalizeAsGs <- function(A, G, hps.to.use=names(A)) {
  maxA <- max(abs(A[hps.to.use]), na.rm=T)
  A <- A / maxA
  G <- G * maxA

  list(A=A, G=G)
}



# computes a consensus depletion profile for a given gene, based on the profiles of its hairpins, by
# fitting the gene data matrix so that x_ij ~= beta_i + A_i * G_j
#
# m - a matrix with depletion data. rows are hps (rownames are hairpin IDs). cols are samples.
# s.values - null distributions of solution fitness scores. used to determine significance of the solution.
# sol.fit.score.method - what method to use to score the goodness of fit of the solution. Solution is accepted or rejected by the p-value for this score, computed using s.values.
# hp.fit.score.method - which method to use to score the agreement of each individual hp with the consensus solution. Lower is better!
# sol.pval.method - the method to use to create a null distribution for sol fit quality.
# min.A.value - the minimum value a hairpin A value can have. Should be positive. Currently using 0.2.
findGeneFit <- function(m, s.values, included.hps=rownames(m), fit.method, sol.fit.score.method,
                        hp.fit.score.method, positive.As=NULL, sol.pval.method, null.significance,
                        sol.accept.method, min.A.value) {
  hp.names <- rownames(m)
  sol.fit.score <- NA
  p.val <- NA
  gene.size <- length(included.hps) # number of hps to use before removing any
  
  while (length(included.hps) >= 2) {
    sol.size <- length(included.hps)
    fit.mat <- m[included.hps,]

    sol <- compConsensus(fit.mat, fit.method, sol.fit.score.method, hp.fit.score.method, positive.As)

    A <- sol$A
    if (sum(!is.na(A)) < 2) {
      browser() #####################
      sol <- NULL # not enough non-missing data points to compute a solution with two hairpins
      break
    }
    
    hps.fit.score <- sol$hps.fit.score
    sol.fit.score <- sol$sol.fit.score
    
    # remove hps with too low efficacy (or negative ones)
    if (min(A, na.rm=T)/max(A, na.rm=T) < min.A.value) {
      hp.to.remove <- included.hps[which.min(A[included.hps])]
    } else { # if all A values are high enough, let the next hp to remove to be the hp with the highest uniqueness
      # the hairpin with the worst fit
      hp.to.remove <- included.hps[which.max(hps.fit.score[included.hps])]

      if (!is.null(s.values)) { # if this is not a run to find the null distribution
        # if the minimization objective (sum of square resid or sum of abs resid) is below the null threshold, we've found a solution
        p.val <- compPvalueFromNull(sol.fit.score, s.values, gene.size, sol.size, sol.pval.method)

        # decide whether to accept or reject the solution
        if (sol.accept.method == 'null.pval') {
          if (p.val < null.significance)
            break
        } else if (sol.accept.method == 'constant') {
          if (sol.fit.score < null.significance)
            break
        } else {
          stop('unsupported sol.accept.method')
        }
      }
    }

    if (is.null(s.values)) # when generating null dist we would like to get the solution using all the input hps.
      break

    included.hps <- setdiff(included.hps, hp.to.remove)
    # cat(paste("Removing hp", hp.to.remove, "with agreement =", hps.agreement[hp.to.remove], "\n"))
  } # main while loop 

  if (!is.null(sol)) {
    if (is.null(sol$weights)) { # used an unweighted method (e.g., "lsfit") so need to define equal weights.
      nh <- nrow(fit.mat)
      ns <- ncol(fit.mat)
      sol$weights <- matrix(1/(nh*ns), nrow=nh, ncol=ns)
    }

    w.resid <- sol$resid * sol$weights # compute weighted residuals
    
    sol$w.resid <- w.resid
    sol$included.hps <- included.hps
    sol$next.hp.to.remove <- hp.to.remove
    sol$hps.agreement <- hps.fit.score[hp.names] # for backward compatibility (ploting the results)
    names(sol$hps.agreement) <- hp.names # for backward compatibility (ploting the results)
    sol$p.val <- p.val
  }
  
  sol
  
  #list(G=G, A=A, beta=beta, hps.agreement=hps.fit.score, included.hps=included.hps, objective=sol$objective, sol.fit.score=sol.fit.score, next.hp.to.remove=hp.to.remove, resid=sol$resid, weights=sol$weights, w.resid=w.resid)
}


# given hps data computes a consensus
compConsensus <- function(fit.mat, fit.method, sol.fit.score.method,
                          hp.fit.score.method, positive.As=NULL) {
  n.hps <- nrow(fit.mat)
  stopifnot(n.hps > 1)

  hp.names <- rownames(fit.mat)
  
  if (fit.method %in% c("lsfit", "l1fit", "wl1fit")) {
    sol <- factanal.ccr(t(fit.mat), factors=1, method=fit.method, print=F, orth=F, drawplot=F)
    G <- sol$scores[,1]
    A <- sol$loadings[,1]
    scale <- sol$scale 
    beta <- sol$center[hp.names]
    names(beta) <- hp.names
  } else if (fit.method %in% c("wl1")) { # my version of weighted l1 fit
    if (n.hps > 2) {
      sol <- wl1.fit.matrix(fit.mat, positive.As)
      G <- sol$G
      A <- sol$A
      scale <- sol$scale
      beta <- sol$mean[hp.names]
    } else { # only 2 hairpins, use lsfit
      sol <- factanal.ccr(t(fit.mat), factors=1, method="lsfit", print=F, orth=F, drawplot=F)
      G <- sol$scores[,1]
      A <- sol$loadings[,1]
      scale <- sol$scale
      beta <- sol$center[hp.names]
      sol$resid <- t(sol$resid) # factanal() outputs the residues in a transposed way.
    }      
    names(beta) <- hp.names
  } else if (fit.method %in% c("l1")) { # my version of un-weighted l1 fit
      sol <- l1.fit.matrix(fit.mat, positive.As)
      G <- sol$G
      A <- sol$A
      names(A) <- hp.names
      scale <- sol$scale
      names(scale) <- hp.names
      beta <- sol$mean[hp.names]
  } else
  stop("Unrecognized fit method")

  # set the sign of the hairpins' A values so that most of them are positive
  switch.sign <- (-1) ^ as.numeric(sum(A<0) > (length(A)/2))
  A <- A[hp.names] * switch.sign # add NAs for hps not included
  names(A) <- hp.names
  G <- G * switch.sign
  scale <- scale[hp.names]
  names(scale) <- hp.names

      
  ### Calculate fitness score for each hairpin ###
  hps.fit.score <- rep(NA, length=length(hp.names))
  names(hps.fit.score) <- hp.names
  if (hp.fit.score.method == "uniqueness") { # better hps have lower uniquness (i.e., lower variation of residues)
    hps.fit.score[hp.names] <- sol$uniquenesses[hp.names] # hps not included in the solution are assigned NA
  } else if (hp.fit.score.method == "spearman.cor") { # better hps have higher spearman correlation with the consensus solution.
    hps.fit.score[hp.names] <- -1 * cor(G, t(fit.mat), method="spearman", use="pairwise.complete.obs") # negate to make lower better.
  } else if (hp.fit.score.method == "MI") { # 1- (NMI(G,hp)/NMI(G,G))
    hps.fit.score[hp.names] <- -1 * mutual.inf.v.mat(G, t(fit.mat))
  }
  else
    stop("Unsupported hp.fit.score.method")

    
  ### compute fitness of solution
  if (sol.fit.score.method == 'sum.resid') {
    sol.fit.score <- sol$objective
  } else if (sol.fit.score.method == 'Rsquare') {
    sol.fit.score <- 1 - getRsquare(fit.mat, sol, fit.method)
  } else if (sol.fit.score.method == 'worst.hp.uniqueness') { # largest MAD of residuals any hp has (equivalent to sum(abs(resid)) )
    sol.fit.score <- max(sol$uniquenesses, na.rm=T)
  } else if (sol.fit.score.method == 'worst.hp.score') { # the worst (highest) fitness score of any hairpin in the solution
    sol.fit.score <- max(hps.fit.score, na.rm=T)
  }
  else
    stop('Unsupported sol.fit.score')

  consensus <- list()
  consensus$A <- A
  consensus$G <- G
  consensus$scale <- scale
  consensus$scaledA <- A * scale
  consensus$MAD <- sol$MAD # mad of original hairpin values 
  consensus$beta <- beta
  consensus$hps.fit.score <- hps.fit.score
  consensus$sol.fit.score <- sol.fit.score
  consensus$objective <- sol$objective
  consensus$resid <- sol$resid
  consensus$pred.m <- sol$pred.m
  consensus$weights <- sol$weights

  consensus
#  list(G=G, A=A, beta=beta, hps.agreement=hps.fit.score, included.hps=included.hps, objective=sol$objective, sol.fit.score=sol.fit.score, next.hp.to.remove=hp.to.remove, resid=sol$resid, weights=sol$weights, w.resid=w.resid)
}


compPvalueFromNull <- function(sol.fit.score, s.values, gene.size, sol.size, sol.pval.method) {
  if (sol.pval.method == 'null.two.levels') {
    nullTable <- s.values[[gene.size]]
    moreExtreme <- sum(nullTable[, sol.size] < sol.fit.score)
#    moreExtreme <- max(moreExtreme, 1) # don't want any 0 p-value
    totalNullPoints <- nrow(nullTable)
    pval <- moreExtreme / totalNullPoints
  } else if (sol.pval.method == 'null.one.level') {
    nullTable <- s.values[[sol.size]]
    moreExtreme <- sum(nullTable[, sol.size] < sol.fit.score)
    totalNullPoints <- nrow(nullTable)
    pval <- moreExtreme / totalNullPoints
  } else {
    stop("unsupported sol.pval.method")
  }

  pval
}

# given the original data and the solution, returns a squared error matrix containing only the hps included in the
# solution.
getSEmatrix <- function(m, included.hps, G, A, beta) {
  data <- m[included.hps,]
  pred <- t(beta + outer(G, A, "*"))
  pred <- pred[included.hps,]

  (data - pred)^2
}


# returns quality measurements for the given solution
obsolete.getSolutionQuality <- function(m, included.hps, G, A, beta) {
  SE <- getSEmatrix(m, included.hps, G, A, beta)
  mean.SE <- list()
  mean.SE$total <- mean(as.numeric(SE), na.rm=T)
  mean.SE$hps <- apply(SE, 1, mean, na.rm=T)
  mean.SE$samples <- apply(SE, 2, mean, na.rm=T)

  median.SE <- list()
  median.SE$total <- median(as.numeric(SE), na.rm=T)
  median.SE$hps <- apply(SE, 1, median, na.rm=T)
  median.SE$samples <- apply(SE, 2, median, na.rm=T)

  list(mean.SE=mean.SE, median.SE=median.SE)
}


# returns quality measurements for a given solution's residuals
getSolutionQuality <- function(resid) {
  l <- list()

  abs.resid <- as.numeric(abs(resid))
  l$abs.resid.sum <- sum(abs.resid, na.rm=T)
  l$abs.resid.mean <- mean(abs.resid, na.rm=T)
  l$abs.resid.median <- median(abs.resid, na.rm=T)
  l$abs.resid.mad <- mad(abs.resid, na.rm=T)

  sqr.resid <- as.numeric(resid^2)
  l$sqr.resid.sum <- sum(sqr.resid, na.rm=T)
  l$sqr.resid.mean <- mean(sqr.resid, na.rm=T)
  l$sqr.resid.median <- median(sqr.resid, na.rm=T)
  l$sqr.resid.mad <- mad(sqr.resid, na.rm=T)

  l
}

# returns one solution using the hairpins specified in hps.to.use. Will use only a subset of the hps.to.use that
# agree enough (according to the values in s.values - significance values).
# solution returned is a list including the fields:
# G - the Essentiality profile
# hps.agreement - a measure of agreement between each hp's profile and the Gs
# included.hps - the hps actually used in the solution.
# A - efficacy values for all included hps
# beta - intercept for each hairpin
getGeneSolution <- function(m, s.values, hps.to.use, fit.method, sol.fit.score.method,
                            hp.fit.score.method, positive.As, sol.pval.method, null.significance,
                            sol.accept.method, min.A.value) {
  if (fit.method == "cor.w.G") { # compute a solution by maximizing sum of corr with hps
    consensus <- findGeneConsensus(m, s.values, hps.to.use) 

    if (length(consensus$included.hps) >= 2) { # if a solution with two or more hairpins exist, continue
      hp.params <- getBestHPparams(m, consensus$G)
      normAG <- normalizeAsGs(hp.params$A, consensus$G, hps.to.use)
      hp.params$A <- normAG$A
      consensus$G <- normAG$G

      # generates a binary string where '1' indicates that this hp was included in the solution. '0' that it wasn't.
      solution.id <- paste(as.character(as.integer(names(hp.params$A) %in% consensus$included.hps)), collapse="")
    
      solution <- c(consensus, hp.params, id=solution.id)
    
    } else {
      solution <- NULL # no solution with at least two hairpins was found
    }
  } # cor.w.G
  else if (fit.method %in% c("lsfit", "l1fit", "wl1fit", "wl1", "l1")) { 
    consensus <- findGeneFit(m, s.values, hps.to.use, fit.method, sol.fit.score.method,
                             hp.fit.score.method, positive.As, sol.pval.method,
                             null.significance=null.significance,
                             sol.accept.method=sol.accept.method,
                             min.A.value=min.A.value) 

    if (!is.null(consensus) && length(consensus$included.hps) >= 2) { # if a solution with two or more hairpins exist, continue
      normAG <- normalizeAsGs(consensus$scaledA, consensus$G, hps.to.use)
      consensus$scaledA <- normAG$A
      consensus$A <- consensus$scaledA / consensus$scale
      consensus$G <- normAG$G

      # generates a binary string where '1' indicates that this hp was included in the solution. '0' that it wasn't.
      solution.id <- paste(as.character(as.integer(rownames(m) %in% consensus$included.hps)), collapse="")
    
      solution <- c(consensus, id=solution.id)
    } else {
      solution <- NULL # no solution with at least two hairpins was found
    }
  } # lsfit, l1fit, wl1fit, wl1, l1
  else {
    stop(paste('Unsupported fit method:', fit.method))
  }

  if (!is.null(solution)) {
    solution$quality <- getSolutionQuality(solution$resid)
    if (!is.null(solution$w.resid)) {
      l <- getSolutionQuality(solution$w.resid)
      names(l) <- gsub('.resid.', '.w.resid.', names(l))
      solution$quality <- c(solution$quality, l)
    }
  }

  
  solution
}


# Returns all the solutions for the given gene.
# output:
# list(
# gene.name
# num.solutions
# solutions - a list containing all the solutions, no name for list elements.
# s.values - list of vectors with cut-offs for solution fitness score, for each size of gene and then for each number
# of hairpins left after removal of non-fitting hps.
getGeneSolutions <- function(gene.name, m, s.values, fit.method="cor.w.G", sol.fit.score.method,
                             hp.fit.score.method, positive.As, sol.pval.method, null.significance,
                             sol.accept.method, min.A.value) {
  solutions <- list()
  hps.to.use <- row.names(m)
  num.hps <- nrow(m)

  # just so we know it's alive.
  if (runif(1) < 0.01) 
    cat('.')

  sol.num <- 1
  while (length(hps.to.use) >= 2) { # repeatedly look for solutions if number of remaining hps is atleast 2
# 10/06/10 - keep the s.values structure all the way down    
#    sol <- getGeneSolution(m, s.values[[length(hps.to.use)]], hps.to.use, fit.method,
#                           sol.fit.score.method, hp.fit.score.method, positive.As)
    sol <- getGeneSolution(m, s.values, hps.to.use, fit.method,
                           sol.fit.score.method, hp.fit.score.method, positive.As, sol.pval.method,
                           null.significance=null.significance,
                           sol.accept.method=sol.accept.method,
                           min.A.value=min.A.value)

    if (is.null(sol)) { # no solution was found
#      cat(paste("No solution for", gene.name, "\n"))
      break
    }

    sol$gene.name <- gene.name
    sol$sol.num <- sol.num
    sol$name <- paste(sol$gene.name, '_', sol$sol.num, '_', sol$id, sep="")
      
    solutions <- c(solutions, list(sol))

    hps.to.use <- setdiff(hps.to.use, sol$included.hps)

    sol.num <- sol.num + 1
  }
  
  list(gene.name=gene.name, num.hps=num.hps, num.solutions=length(solutions), solutions=solutions)
}


# Returns the solutions for all the genes in cons.gene.map or those specified in gene.list
#
# m - data matrix. hps in rows with row.names=hp ID.
# cons.gene.map - vector the size of nrow(m) with values signifying the gene name for each hp.
# gene.list

getGenesSolutions <- function(m, cons.gene.map, s.values, gene.list=unique(cons.gene.map),
                              num.cores=1, fit.method, sol.fit.score.method, hp.fit.score.method, positive.As,
                              sol.pval.method, null.significance, sol.accept.method, min.A.value) {
  if (is.null(gene.list))
    gene.list <- unique(cons.gene.map)
  
  if (num.cores > 1) {
    require(multicore, quietly=TRUE, warn.conflicts=FALSE)
    l <- mclapply(gene.list, function(g) getGeneSolutions(g,
                                                          m[cons.gene.map==g,],
                                                          s.values,
                                                          fit.method,
                                                          sol.fit.score.method, hp.fit.score.method,
                                                          positive.As, sol.pval.method,
                                                          null.significance=null.significance,
                                                          sol.accept.method=sol.accept.method,
                                                          min.A.value=min.A.value),
                  mc.cores=num.cores)
  } else {
    l <- lapply(gene.list, function(g) getGeneSolutions(g,
                                                        m[cons.gene.map==g,],
                                                        s.values,
                                                        fit.method,
                                                        sol.fit.score.method, hp.fit.score.method,
                                                        positive.As, sol.pval.method,
                                                        null.significance=null.significance,
                                                        sol.accept.method=sol.accept.method,
                                                        min.A.value=min.A.value))
  }
  
  names(l) <- lapply(l, function(x) x$gene.name)
  
  l
}


# Generates a HP-HP correlation matrix for each gene in the givem gene.list.
# Correlation method used is biwt.corr() of the biwt package. This takes time...
generateGenesHPHPcor <- function(gene.list, m, cons.gene.map, num.cores=1) {

  if (is.null(gene.list))
    gene.list <- unique(cons.gene.map)

  inner <- function(gene.name, m, cons.gene.map) {
    # Generate a hp-hp correlation data for this gene
    require(biwt, quietly=TRUE, warn.conflicts=FALSE)
    hps.to.use <- which(cons.gene.map == gene.name)
    gene.data <- m[hps.to.use,]
    hphp.cor <- biwt.cor(gene.data, r=0.2, output='matrix')
    rownames(hphp.cor) <- rownames(gene.data)
    colnames(hphp.cor) <- rownames(gene.data)
    
    hphp.cor
  }
  
  if (num.cores > 1) {
    require(multicore, quietly=TRUE, warn.conflicts=FALSE)
    l <- mclapply(gene.list, function(g) inner(g, m, cons.gene.map), mc.cores=num.cores)
  } else {
    l <- lapply(gene.list, function(g) inner(g, m, cons.gene.map))
  }

  names(l) <- gene.list
  
  l
}



# generates plots describing the given solution (this is one solution for the given gene)
plotSolution <- function(sol.name, gene.name, m, solution, s.values, results=NULL) {
  require(gplots, quietly=TRUE, warn.conflicts=FALSE)
  require(RColorBrewer, quietly=TRUE, warn.conflicts=FALSE)

  hp.IDs <- names(solution$hps.agreement)
  gene.size <- length(hp.IDs)
  num.included.hps <- length(solution$included.hps)
  asterisk <- ifelse(hp.IDs %in% solution$included.hps, "*", "")
  hp.names <- paste(asterisk, substring(hp.IDs,9,14), sep="")
#  max.eff.hp <- names(which.max(solution$A)) # hp ID of hp with max efficacy
  most.negative.hp <- names(which.min(apply(m[solution$included.hps,], 1, median, na.rm=T)))

  ###  page 1  ###
  par(mfrow=c(3,1), cex.axis=0.7)
  sample.names <- sub("_.*","", colnames(m))
  sample.lineages <- sub("^[a-zA-Z0-9]*_","", colnames(m))
  unique.lineages <- unique(sample.lineages)
  bar.colors <- brewer.pal(12,"Set3")[apply(outer(unique.lineages, sample.lineages, '=='), 2, which) %% 12 + 1]# pure craziness 
  sample.order <- order(sample.lineages, sample.names)
  # relative essentiality barplot
  barplot(solution$G[sample.order],
          names=sample.names[sample.order],
          main=paste(sol.name,"Relative Essentiality"),
          col=bar.colors[sample.order], las=2)

  # hairpin qualities barplot
  maxA <- max(solution$A[hp.IDs], na.rm=T)
  barplot(rbind(solution$A[hp.IDs] / maxA, solution$MAD[hp.IDs], -solution$hps.agreement[hp.IDs]), beside=T, col=c("dark red", "dark green", "cyan"),
            names=hp.names, ylim=c(0,1), legend.text=c("A (scaled 0-1)", "Hairpin MAD", "Hairpin fit score"),
#            main=paste(sol.name, "HP Efficacy and agreement w/ consensus"), sub=paste("solution fit score is", round(solution$sol.fit.score,2), "(cut-off is", round(s.values[[gene.size]][num.included.hps],2),")") )
            main=paste(sol.name, "HP Efficacy and agreement w/ consensus"), sub=paste("solution fit score is", round(solution$sol.fit.score,2), "(p-val =", round(solution$p.val, 2),")") )

  abline(h=0)
#  abline(a=s.values[[gene.size]][num.included.hps], b=0, lty=21) # hps have to have lower MAD of the residuals than this

  # location of hairpins           
  barplot(rbind(solution$beta[hp.IDs], apply(m[hp.IDs,], 1, mean), apply(m[hp.IDs,], 1, median)), beside=T, names=hp.names,
            main=paste("beta, mean, median values for all hps"))

  
  ### page 2 ###
  par(mfrow=c(3,1), cex.axis=0.7)
  barplot(solution$G, names=colnames(m), main=paste(sol.name,"Relative Essentiality"), col="navy", las=2)

  # 2010/2/11 adding a plot showing gene essentialities from ZMADed Gs matrix.
  # Added for the analysis of Melanomas (David K)
  barplot(as.numeric(results$GsZMAD[sol.name,]), names=colnames(m), main=paste(sol.name,"ZMADed Relative Essentiality"), col="red", las=2)
  
  
#  est.max.eff.hp.profile <- solution$beta[max.eff.hp] + solution$G
#  barplot(est.max.eff.hp.profile, names=colnames(m),
#          main=paste(sol.name,"Estimated profile of most efficacious hp (", max.eff.hp," )"), col="navy", las=2)

  # plot profile of most negative hp that is included in the analysis
   barplot(m[most.negative.hp,], names=colnames(m),
           main=paste(sol.name,"Estimated profile of most negative hp (", most.negative.hp," )"), col="navy", las=2)

 
  ### page 3 ###
  # plot correlation pairs
  plotCorrPairs(m[hp.IDs,], paste("Correlations of", gene.name, "'s hairpins"), results$gene.cor.pvals[[gene.name]])

if (0) { ### skip the rest ##### 2011/5/1
  ### pages 4-7 ###
  # now for some heatmaps
  hmcol <- colorRampPalette(rev(brewer.pal(10, "RdBu")))(256)

  data <- m[hp.IDs, ]
  ylim <- c(min(data, na.rm=T), max(data, na.rm=T))
  # original data
  data.to.plot <- rbind(solution$G, as.matrix(data))
  heatmap.2(data.to.plot, labCol=colnames(m), labRow=c("consensus", hp.names),
            main=paste(gene.name,"original values"), col=hmcol, trace="row",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", breaks=seq(-4,4,8/256))
  
  # row-normalized  
  heatmap.2(data.to.plot, labCol=colnames(m), labRow=c("consensus", hp.names),
            main=paste(gene.name,"row-normalized values"), col=hmcol, trace="row",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", symkey=F, scale="row", breaks=seq(-4,4,8/256))

  # row-normalized, sorted by consensus values
  heatmap.2(data.to.plot[,order(solution$G)], labCol=colnames(m)[order(solution$G)], labRow=c("consensus", hp.names),
            main=paste(gene.name,"row-normalized values\nSamples ordered by consensus profile"), col=hmcol, trace="row",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, scale="row", na.color="black", symkey=F, breaks=seq(-4,4,8/256))

  if (0) { #12/10/2010
  # row-normalized clustered
  heatmap.2(data.to.plot, labCol=colnames(m), labRow=c("consensus", hp.names),
            main=paste(gene.name,"row-normalized values, hps clustered"), col=hmcol, trace="row",
            Rowv=T, Colv=F, dendrogram="row", cex.axis=0.6, scale="row", na.color="black", symkey=F,
            distfun=function(x) as.dist(1-cor(t(x), method="spearman", use="na.or.complete")), breaks=seq(-4,4,8/256))
}
  

  if (!is.null(solution$resid)) {
    res.col <- colorRampPalette(brewer.pal(11, "RdGy"))(256)
    heatmap.2(solution$resid, labCol=colnames(m), labRow=rownames(solution$resid),
              main=paste(sol.name,"residuals"), col=res.col, trace="none",
                Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, breaks=seq(-2,2,4/256))
  }

  # the below gives an error of unknown source... 07/21/10
#  if (!is.null(solution$weights)) {
#    res.col <- colorRampPalette(brewer.pal(11, "RdGy"))(256)
#    heatmap.2(solution$weights, labCol=colnames(m), labRow=rownames(solution$weights),
#              main=paste(sol.name,"weights"), col=res.col, trace="none",
#              Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, breaks=seq(0, 0.01,0.01/256))
#  }

  if (0) { # 2010/11/11 nobody cares about this one
    if (!is.null(solution$w.resid)) {
      res.col <- colorRampPalette(brewer.pal(11, "RdGy"))(256)
      heatmap.2(solution$w.resid, labCol=colnames(m), labRow=rownames(solution$w.resid),
                main=paste(sol.name,"weighted residuals"), col=res.col, trace="none",
                Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, cex.lab=0.6, breaks=seq(-0.01,0.01,0.02/256))
    }
  }


if (0) {
  # row-normalized clustered, sorted by consensus values
  heatmap.2(data.to.plot[,order(solution$G)], labCol=colnames(m)[order(solution$G)], labRow=c("consensus", hp.names),
            main=paste(gene.name,"row-normalized values, hps clustered\nSamples ordered by consensus profile"), col=hmcol, trace="row",
            Rowv=T, Colv=F, dendrogram="row", cex.axis=0.6, scale="row", na.color="black", symkey=F,
            distfun=function(x) as.dist(1-cor(t(x), method="pearson", use="na.or.complete")))
} # if (0)

  
  if (!is.null(results))
    plotMostCorrGenesToGene(results, sol.name)

} # if(0) skipping all pages but 1-3
}

# generates PDF files, one for each gene solution.
#
# solutions - a list with the gene name as names of elements. as generated by getGenesSolutions
generateSolutionPlots <- function(m, s.values, all.genes.solutions, output.dir='./',
                                  gene.list=NULL, num.cores=1, results=NULL, nullGGCorr=NULL) {
  if (is.null(gene.list))
    gene.list=names(all.genes.solutions)

  cat(paste("Gene list has", length(gene.list), "genes in it\n"))
  
  inner <- function(output.dir, m, s.values, gene.solutions) {
    num.solutions <- gene.solutions$num.solutions # how many different solutions does this gene have?

    if (!is.null(num.solutions) && num.solutions > 0) {
    # each solution deserves a separate file
      for (s in gene.solutions$solutions) {
        gene.name <- gene.solutions$gene.name
#        s.name <-  paste(gene.name, '_', s$id, sep="")
        s.name <- s$name
        file.name <- paste(output.dir, s.name, '.plots.pdf', sep="")
        
        pdf(height=11, width=8.5, file=file.name)
        plotSolution(s.name, gene.name, m, s, s.values, results)      
        dev.off()
        generateMostCorrToSol(s.name, results, output.dir, nullGGCorr=nullGGCorr)
      }
    }
  } # inner

  if (num.cores > 1) {
    require(multicore, quietly=TRUE, warn.conflicts=FALSE)
    l <- mclapply(gene.list, function(g) inner(output.dir, m, s.values, all.genes.solutions[[g]]), mc.cores=num.cores)
  } else {
    l <- lapply(gene.list, function(g) inner(output.dir, m, s.values, all.genes.solutions[[g]]))
  }

  l
}


#  res <- runAnalysis("fit_55k_091206")
# generateSolutionPlots(res$m, res$s.values, res$gene.solutions, './', c("KRAS", "STK33"))

# returns solutions for num.samples bootstraps of the data for a given gene.
generateGeneBS <- function(m, cons.gene.map, s.values, geneID, num.samples=100, num.cores=1,
                           fit.method, sol.fit.score.method, hp.fit.score.method, positive.As,
                           sol.pval.method, min.A.value) {
  samples <- generateBSsamples(ncol(m), num.samples)
  
  if (num.cores > 1) {
    require(multicore, quietly=TRUE, warn.conflicts=FALSE)
    l <- mclapply(1:num.samples, function(s) getGeneSolutions(geneID,
                                                              m[cons.gene.map==geneID, samples[s,]],
                                                              s.values,
                                                              fit.method,
                                                              sol.fit.score.method, hp.fit.score.method,
                                                              positive.As, sol.pval.method, sol.accept.method,
                                                              min.A.value=min.A.value),
                  mc.cores=num.cores)
  } else {
    l <- lapply(1:num.samples, function(s) getGeneSolutions(geneID,
                                                            m[cons.gene.map==geneID, samples[s,]],
                                                            s.values,
                                                            fit.method,
                                                            sol.fit.score.method, hp.fit.score.method,
                                                            positive.As, sol.pval.method, sol.accept.method,
                                                            min.A.value=min.A.value))
  }
  
  l
}



# generates bootstrap sample IDs. (num.samples samples each of size num.elements)
generateBSsamples <- function(num.elements, num.samples) {
  matrix(sample(num.elements, num.samples*num.elements, replace=T), nrow=num.samples)
}


checkIncludedHPs <- function(l) {

  included.hps <- unlist(lapply(1:length(l), function(i) {
    tmp <- l[[i]]$solutions[[1]]
     paste(as.character(as.numeric(names(tmp$hps.agreement) %in% tmp$included.hps)), collapse="")

  }))

  sort(table(included.hps), decreasing=T)
}


BSgene <- function(results, geneID, num.samples=1000, num.cores=12, fit.method="cor.w.G") {
  l <- generateGeneBS(results$m, results$cons.gene.map, results$s.values, geneID, num.samples, num.cores, fit.method)
  checkIncludedHPs(l)
}



# returns a 3D array, with results of corr between rows of m  with bootstrapped columns, num.samples times.
# first coordinate of the array is the sample number. the other two are hp.
generateBScorr <- function(m, num.samples=100) {
  corrs <- array(0, dim=c(num.samples, nrow(m), nrow(m)))
  samples <- generateBSsamples(ncol(m), num.samples)

  for (i in 1:num.samples) {
    corrs[i, ,] <- as.matrix(cor(t(m[,samples[i,]]), use='na.or.complete'))
  }

  corrs
}


# for a given, fixed, hps of a gene. Find best Gs, As, betas across bootstrap samples.
#
# BS.idx - matrix with each row containing the indices of m's columns comprising one BS sample.
getBSstability <- function(m, BS.idx) {
  num.BSsamples <- nrow(BS.idx)
  num.hps <- nrow(m)
  num.samples <- ncol(m)
  
  Gs <- matrix(0, nrow=num.BSsamples, ncol=num.samples)
  As <- matrix(0, nrow=num.BSsamples, ncol=num.hps)
  betas <- matrix(0, nrow=num.BSsamples, ncol=num.hps)
  agreements <-  matrix(0, nrow=num.BSsamples, ncol=num.hps)
  
  for (s in 1:num.BSsamples) {
    mBS <- m[, BS.idx[s,]]
    G <- getMaxSumCorVector(mBS)
    hps.agreement <- getHPsAgreement(mBS, G)
    hp.params <- getBestHPparams(mBS, G)
    normAG <- normalizeAsGs(hp.params$A, G)
    
    As[s,] <- normAG$A
    Gs[s,] <- normAG$G # doesn't mean anything as Gs are bootstrapped!!!
    betas[s,] <- hp.params$beta
    agreements[s,] <- hps.agreement
  }

  list(As=As, Gs=Gs, betas=betas, agreements=agreements)
}



getGeneSolutionBSstability <- function(results, geneID, num.BSsamples=200) {
  BS.idx <- generateBSsamples(ncol(results$m), num.BSsamples)
  getBSstability(results$m[results$gene.solutions[[geneID]]$solutions[[1]]$included.hps,], BS.idx)
}


# given the results, extract a table with a row for each gene, with information about the solutions
# row.names - gene name
# num.solutions - number of solutions generated for this gene
# num.hps - number of hairpins data we have for this gene
extractGeneTable <- function(results) {
  sols <- results$gene.solutions
  gnames <- vector(mode="character", length=length(sols))
  nsol <- vector(mode="numeric", length=length(sols))
  nhps <- vector(mode="numeric", length=length(sols))
  
  for (i in 1:length(sols)) {
    gnames[i] <- sols[[i]]$gene.name
    nsol[i] <- sols[[i]]$num.solutions
    nhps[i] <- sols[[i]]$num.hps
  }

  gtable <- data.frame(num.solutions=nsol, num.hps=nhps, row.names=gnames)

}


# extract a matrix from the results with each row containing an essentiality profile of each solution.
# rownames - solution name.
# column for each sample.
extractGTable <- function(results, num.solutions) {
  Gs <- matrix(nrow=num.solutions, ncol=results$num.samples)
  sol.names <- NULL
  r <- 1 # current row
  for (x in results$gene.solutions) {
    ### TODO: no need for checking how many solutions are there for this gene. the name of each solution should use its solution$id property and that's it.
    if (x$num.solutions > 0) {
      for (s in x$solutions) {
#        s.name <-  paste(x$gene.name, '_', s$id, sep="")
        s.name <- s$name
        Gs[r,] <- s$G
        sol.names <- c(sol.names, s.name)
        r <- r+1
      }
      
#      m <- matrix(0, nrow=x$num.solutions, ncol=length(x$solutions[[1]]$G))
#      if (x$num.solutions > 1) {
#        rownames(m) <- paste(x$gene.name, '_', 1:x$num.solutions, sep="")
#      } else {
#        rownames(m) <- x$gene.name
#      }
      
#      for (i in 1:x$num.solutions) {
#        m[i,] <- x$solutions[[i]]$G
#      }
#      Gs[r:(r+nrow(m)-1),] <- m
#      # TODO s.name <- paste(x$gene.name, '_', x$solutions[[s]]$id, sep="")
#      sol.names <- c(sol.names, rownames(m))
#      r <- r + nrow(m)
    }
  }
  colnames(Gs) <- colnames(results$m)
  rownames(Gs) <- sol.names
  as.data.frame(Gs)
}


# given the results, extracts a table with a row for each solution.
# rownames - solution name
# gene.name - gene name
# num.hps - number of hairpins used in this solution.
# mean.error - mean prediction error for all the data points of this solution's hairpins (excluding NAs).
# median.error
extractSolutionTable <- function(results) {
  for (s in results$gene.solutions) {
    if (s$num.solutions > 0) {
      quality.col.names <- names(s$solutions[[1]]$quality)
      break
    }
  }
  
  col.names <- c("gene.name", "num.hps", quality.col.names)
  num.cols <- length(col.names)

  sols <- as.data.frame(matrix(nrow=results$total.num.solutions, ncol=num.cols,
                 dimnames=list(NULL, col.names)))
#  sol.names <- NULL
  r <- 1 # current row
  for (x in results$gene.solutions) {
    if (x$num.solutions > 0) {
      for (s in x$solutions) {
#        s.name <-  paste(x$gene.name, '_', s$id, sep="")
        s.name <- s$name
        sols[r,"num.hps"] <- length(s$included.hps)
        sols[r,"gene.name"] <- x$gene.name
        sols[r, quality.col.names] <- unlist(s$quality[quality.col.names])
        rownames(sols)[r] <- s.name
        r <- r+1
        if (r %% 100 == 0) {
          cat('.')
        }
      } # s in x$solutions
    } # of x$num.solutions > 0
  } # x in results$gene.solutions
  
  sols
}


# extracts a table with solutions in rows and samples in columns, with quality measurements as elements.
# currently reports one table with mean.RMSE and another with median.RMSE
extractQualityTables <- function(results) {
  num.samples <- ncol(results$m)
  mean.SE <- matrix(nrow=results$total.num.solutions, ncol=num.samples,
                 dimnames=list(rownames(results$solutions), colnames(results$m)))
  median.SE <- matrix(nrow=results$total.num.solutions, ncol=num.samples,
                 dimnames=list(rownames(results$solutions), colnames(results$m)))
  
  for (x in results$gene.solutions) {
### TODO: no need for checking how many solutions are there for this gene. the name of each solution should use its solution$id property and that's it.

    if (x$num.solutions > 0) {
      for (s in x$solutions) {
#        s.name <-  paste(x$gene.name, '_', s$id, sep="")
        s.name <- s$name
        mean.SE[s.name,] <- s$quality$mean.SE$samples
        median.SE[s.name,] <- s$quality$median.SE$samples
      }

#      for (i in 1:x$num.solutions) {
#        # determine the name of the solution
#        if (x$num.solutions > 1) {
#          sol.name <- paste(x$gene.name, '_', i, sep="")
#        } else {
#          sol.name <- x$gene.name
#        }
        
#        mean.SE[sol.name,] <- x$solutions[[i]]$quality$mean.SE$samples
#        median.SE[sol.name,] <- x$solutions[[i]]$quality$median.SE$samples
#      } # all solutions for a gene
    }
  } # all genes
  quality <- list(mean.SE=list(samples=mean.SE), median.SE=list(samples=median.SE))
}


# generates cutoffs for the solution fitness score for each size of gene to analyze and each number of hps used
# in a potential solution
# uses in params: genes.to.analyze, min.num.null.points, num.null.points, s.values.file, identifier, null.significance,
# num.cores, fit.method, sol.fit.score.method, hp.fit.score.method, sol.accept.method, const.cutoff.
generateSignificanceCutoffs <- function(params, m, cons.gene.map) {
  
  # compute null hp agreement thresholds
  if (is.null(params$genes.to.analyze)) {
    cons.to.analyze <- cons.gene.map # analyze all data
  } else {
    cons.to.analyze <- cons.gene.map[cons.gene.map %in% params$genes.to.analyze]
  }
  
  gene.size.count <- table(table(cons.to.analyze)) # count of number of genes present in the data of each size
  gene.sizes <- 2:max(unique(table(cons.to.analyze))) # we have to create null distributions for genes of these sizes 

  
  if (is.null(params$s.values.file)) {
    s.values.file <- paste(params$results.dir, params$identifier, '.s.values.Rdata', sep="")
  } else {
    s.values.file <- params$s.values.file
  }
  
  if (file.exists(s.values.file)) {
    cat(paste("Loading s.values from file:", s.values.file, "\n"))
    load(s.values.file)
  } else {
    cat(paste("Generating s.values and saving to file:", s.values.file, "...\n"))
    # for all gene sizes we need to generate at least min.num.null.points values from random genes
    num.null.points <- rep(params$min.num.null.points, max(gene.sizes))
    # for each gene size use a different number of null points,
    # relative to the number of genes of that size represented in the data.
    num.null.points[as.numeric(names(gene.size.count))] <-
      ceiling(unlist(lapply(gene.size.count / max(gene.size.count) *
                            params$num.null.points, max, params$min.num.null.points)))
      
    s.values <- getSignificanceCutoff(m, num.null.points, gene.sizes,
                                      null.significance=params$null.significance, num.cores=params$num.cores,
                                      fit.method=params$fit.method, sol.fit.score.method=params$sol.fit.score.method,
                                      params$hp.fit.score.method, params$positive.As, params$sol.pval.method,
                                      sol.accept.method=params$sol.accept.method,
                                      min.A.value=params$min.A.value)
    
#    save(s.values, file=s.values.file)
  } # generating s.values, not loading from file

  s.values
}


# returns a list of the most correlated Gs to a given gene's profile
mostCorrToGene <- function(Gs, geneID, num=10, method="spearman") {
  profile <- Gs[geneID,]
  cors <- cor(t(profile), t(Gs),  use="pairwise.complete.obs", method=method)
  ord <- order(cors, decreasing=T, na.last=T)

  cors[,ord[1:num]]
}


# prints a list of the most correlated Gs to a given profile
mostCorrToProfile <- function(Gs, profile, num=10) {
  cors <- cor(profile, t(Gs),  use="na.or.complete", method="spearman")
  ord <- order(cors, decreasing=T)

  cors[,ord[1:num]]
}



# plots a pairs plot showing correlation structure of hairpins of a gene
plotCorrPairs <- function(m, title="", cor.pvals=NULL) {

  panel.cor <- function(x, y, digits=2, prefix="", cex.cor)
    {
      usr <- par("usr"); on.exit(par(usr))
      par(usr = c(0, 1, 0, 1))
      r <- cor(x, y, use='na.or.complete', method='spearman')
      txt <- format(c(r, 0.123456789), digits=digits)[1]
      txt <- paste(prefix, txt, sep="")
      if(missing(cex.cor)) cex.cor <- 0.8/strwidth(txt)
      text(0.5, 0.5, txt, cex = cex.cor * abs(r), col="orange")
      
    }
  
  

 panel.cor.pvalue <- function(x, y, digits=1, prefix="", cex.cor, cor.pvals, org.values)
    {
      usr <- par("usr"); on.exit(par(usr))
      par(usr = c(0, 1, 0, 1))

      if (is.null(cor.pvals)) {
        coorToUse <- !is.na(x) & !is.na(y)
        if (sum(coorToUse) < 3) {
          r <- NA
        } else {
          r <- cor.test(x, y, alternative='g', na.action='na.omit', method='spearman')$p.value
          if (r > 0)
            r <- -log10(r)
          else
            r <- -log10(2.2e-16)
        }
      } else { # we are given the pval to plot already
        x.id <- which(apply(x == org.values, 2, all))
        y.id <- which(apply(y == org.values, 2, all))
        r <- -log10(cor.pvals[x.id, y.id])
      }
      
      txt <- format(c(r, 0.123456789), digits=digits)[1]
      txt <- paste(prefix, txt, sep="")
      if(missing(cex.cor)) cex.cor <- 0.8/strwidth(txt)
      size <- min(3, r, na.rm=T) / 3
      
      text(0.5, 0.5, txt, cex = cex.cor * size, col="blue")
      
    }
  
  # plots a density of bootstrapped correlation values between x and y
  panel.cor.BS.dist <- function(x, y, num.samples=500, ...) {
    corrs <- vector(mode="numeric", length=num.samples)
    for (i in 1:num.samples) {
      s <- sample(length(x), replace=T)
      corrs[i] <- cor(x[s], y[s], use='na.or.complete', method='pearson')
    }
    
    usr <- par("usr"); on.exit(par(usr))
    par(usr = c(-1, 1, -0.1, 3) )    
    lines(density(corrs, from=-1, to=1))
    abline(v=0)
    m <- median(corrs)
    if (sum(corrs>0)/length(corrs) > 0.95)
      col <- 'red'
    else if (sum(corrs<0)/length(corrs) > 0.95)
      col <- 'blue'
    else
      col <- 'grey'
    
    abline(v=m, col=col)

  }
  
  panel.hist <- function(x, ...)
    {
      usr <- par("usr"); on.exit(par(usr))
      par(usr = c(usr[1:2], 0, 1.5) )
      h <- hist(x, plot = FALSE)
      breaks <- h$breaks; nB <- length(breaks)
      y <- h$counts; y <- y/max(y)
      rect(breaks[-nB], 0, breaks[-1], y, col="cyan", ...)
    }
  
  old.panel.barplot <- function(x, ...)
    {
      usr <- par("usr"); on.exit(par(usr))
                                        #    par(usr = c(0, length(x), 0, max(x)-min(x)+0.5) )
                                        #    rect(0:(length(x)-1), 0, 1:length(x), x-min(x)+0.1, col="lightgray")
      if (sum(is.na(x)) != 14) {
        x <- x - mean(x, na.rm=T)
        
        
        par(usr = c(0, length(x), min(x, na.rm=T), max(x, na.rm=T)+0.5) )
        rect(0:(length(x)-1), 0, 1:length(x), x, col="cyan")
      }
    }

   panel.barplot <- function(x, ...)
    {
      usr <- par("usr"); on.exit(par(usr)) 
      par(usr = c(-1, length(x)+1, -2, 2+0.5) )
      lines(x, type='h')
   }

#  panel.points <- function(x, y, ...) {
#    points(x, y, col="red", pch=19, ...)
#  }
  
  panel.points <- function(x, y, ...) {
    panel.smooth(x, y, col.smooth='red', col='dark green', ...)
  }

  labels <- substr(rownames(m), 1, 14)
  plot.data <- t(as.matrix(m))
  pairs(x=plot.data, labels=labels, lower.panel=panel.points,  upper.panel=panel.cor.pvalue, main=title, diag.panel=panel.barplot, cor.pvals=cor.pvals, org.values=plot.data)
  
#  pairs(x=t(as.matrix(m)), labels=rownames(m), lower.panel=panel.points,  upper.panel=panel.cor.BS.dist, main=title, diag.panel=panel.barplot)

}


  
plotMostCorrGenesToGene <- function(results, geneID, num.pair.genes=10, num.heatmap.genes=25) {
  if (dim(results$solutions)[1] < num.pair.genes) {
    cat(paste('not plotting most correlated genes to',geneID,'due to low number of solutions found.\n'))
    return(-1)
  }
    
  genesToPlot <- names(mostCorrToGene(results$Gs, geneID, max(num.pair.genes, num.heatmap.genes)))
  m <- results$Gs[genesToPlot[1:num.pair.genes],]
  plotCorrPairs(m, paste("Genes with most similar essentiality profile to", geneID))

  genesToPlot.zmad <- names(mostCorrToGene(results$GsZMAD, geneID, max(num.pair.genes, num.heatmap.genes)))
  m <- results$GsZMAD[genesToPlot.zmad[1:num.pair.genes],]
  plotCorrPairs(m, paste("Genes with most similar ZMADed essentiality profile to", geneID))

  require(gplots, quietly=TRUE, warn.conflicts=FALSE)
  require(RColorBrewer, quietly=TRUE, warn.conflicts=FALSE)
  hmcol <- colorRampPalette(rev(brewer.pal(10, "RdBu")))(256)

  m <- results$Gs[genesToPlot[1:num.heatmap.genes],]

  data.to.plot <- as.matrix(m)
  heatmap.2(data.to.plot, labCol=colnames(m), labRow=rownames(m),
            main=paste("Genes with essentiality profiles similar to", geneID), col=hmcol, trace="none",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", symkey=F)
  

  data.to.plot <- as.matrix(m[,order(m[1,])]) # order all profiles by the values of the first one (which should be the target gene) 
  heatmap.2(data.to.plot, labCol=colnames(data.to.plot), labRow=rownames(m),
            main=paste("Genes with essentiality profiles similar to", geneID), col=hmcol, trace="none",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", symkey=F)

  
  m <- results$GsZMAD[genesToPlot.zmad[1:num.heatmap.genes],]

  data.to.plot <- as.matrix(m)
  heatmap.2(data.to.plot, labCol=colnames(m), labRow=rownames(m),
            main=paste("Genes with ZMADed essentiality profiles similar to", geneID), col=hmcol, trace="none",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", symkey=F)
  

  data.to.plot <- as.matrix(m[,order(m[1,])]) # order all profiles by the values of the first one (which should be the target gene) 
  heatmap.2(data.to.plot, labCol=colnames(data.to.plot), labRow=rownames(m),
            main=paste("Genes with ZMADed essentiality profiles similar to", geneID), col=hmcol, trace="none",
            Rowv=F, Colv=F, dendrogram="none", cex.axis=0.6, rowsep=1, na.color="black", symkey=F)

}


writeSummary <- function(results, results.dir, identifier) {
  op <- options()
  options(digits=3)
  
  out.file <- paste(results.dir, identifier, '.results.summary.txt', sep="")
  write("Summary of analysis:", file=out.file)

  write(paste("Num of solutions created:", nrow(results$solutions)), file=out.file, append=T)
  
  genes.w.sol <- sum(results$gene.table[,"num.solutions"]>0)
  tot.genes <- nrow(results$gene.table)
  write(paste("Num of genes with solutions:", genes.w.sol, "/", tot.genes, "(", genes.w.sol/tot.genes, ")"), file=out.file, append=T)

  num.used.hps <- sum(results$solutions[,"num.hps"])
  tot.hps <- sum(results$gene.table[,"num.hps"])
  write(paste("Num of hairpins used:", num.used.hps, "/", tot.hps, "(", num.used.hps/tot.hps, ")"), file=out.file, append=T)
  
  write("\nNum of solutions by gene size and by number of solutions per gene:", file=out.file, append=T)
  tmp <- table(results$gene.table)
  tmp <- rbind(colnames(tmp), tmp)
  tmp <- format(cbind(rownames(tmp), tmp), width=6)
  write.table(tmp, file=out.file, append=T, sep='\t', quote=F, row.names=F, col.names=F)

  
  write("\nNum of solutions by solution size:", file=out.file, append=T)
  tmp <- table(results$solutions[, "num.hps"])
  tmp <- rbind(names(tmp), tmp)
  tmp <- format(cbind(rownames(tmp), tmp), width=6)
  write.table(tmp, file=out.file, append=T, sep='\t', quote=F, row.names=F, col.names=F)

  if (0) {
  solution.sizes <- results$solutions[,"num.hps"]
  abs.resid.mean.by.solution.size <- split(results$solutions[,"abs.resid.mean"], solution.sizes)
  tmp <- sapply(abs.resid.mean.by.solution.size, summary) # creates a table with a summary for
  tmp <- rbind(colnames(tmp), tmp)
  tmp <- format(cbind(rownames(tmp), tmp), width=6)
  write("\nStatistics of abs.resid.mean across all solutions, by solution size:", file=out.file, append=T)
  write.table(tmp, file=out.file, append=T, sep='\t', quote=F, row.names=F, col.names=F)
  
  abs.w.resid.sum.by.solution.size <- split(results$solutions[,"abs.w.resid.sum"], solution.sizes)
  tmp <- sapply(abs.w.resid.sum.by.solution.size, summary) # creates a table with a summary for
  tmp <- rbind(colnames(tmp), tmp)
  tmp <- format(cbind(rownames(tmp), tmp), width=6)
  write("\nStatistics of abs.w.resid.sum across all solutions, by solution size:", file=out.file, append=T)
  write.table(tmp, file=out.file, append=T, sep='\t', quote=F, row.names=F, col.names=F)
} # if (0)
  
#  mean.mean.SE <- mean(results$solutions[,"abs.resid.mean"])
#  mean.median.SE <- mean(results$solutions[,"abs.w.resid.sum"])
#  write(paste("Mean of abs.w.resid.sum across all solutions:", mean.median.SE), file=out.file, append=T)

#  write("Median Root-Median-Square-Error for each sample across the gene solutions:", append=T, file=out.file)
#  meds <- sort(apply(results$quality$median.SE$samples, 2, median, na.rm=T))
#  for (s in names(meds)) {
#    write(paste(s, '\t', meds[s], sep=""), file=out.file, append=T)
#  }
  options(op)
}





# generates a list of most correlated solutions to a given solution.
# computes p-values, FDR.
generateMostCorrToSol <-  function(sol.name, results, output.dir, num.top=500, nullGGCorr) {
  n.sol <- nrow(results$Gs)
  num.top <- min(num.top, n.sol-1)
  corrs <- mostCorrToGene(results$Gs, sol.name, num.top)
  gene.names <- unlist(lapply(strsplit(names(corrs),'_'), function(x) x[[1]]))
  pvals <- apply(as.matrix(corrs), 1, function(x) { (sum(nullGGCorr > x, na.rm=T) + 1)  }) / (sum(!is.na(nullGGCorr)) + 1)
  adj.pval <- p.adjust(p=pvals, method="fdr", n=n.sol)
  
  top <- cbind(1:num.top, names(corrs)[1:num.top], gene.names[1:num.top], signif(corrs[1:num.top], 5),
               signif(pvals[1:num.top], 5),
               signif(adj.pval[1:num.top], 5))
  colnames(top) <- c("rank", "solutionID", "GeneID", "correlation", "p-value", "FDR")
  
  file.name <- paste(output.dir, sol.name, '.top', num.top, '.txt', sep='')
  write.table(top, quote=F, file=file.name, sep='\t', row.names=F)
}


# extracts from the results a table with the following columns:
# shRNA ID
# Gene symbol
# isUsed - Participates in a solution? (boolean)
# solution number - (1, 2, ...)
# solution full name  - e.g. GENE_2_010011 (so we can get the participating shRNAs easily from solution full name)
# solution id - 010010
# c-score/p-value/q-value
extractHairpinTable <- function(results) {

  genes <- unique(names(results$gene.solutions)) # genes to process are only those that were analyzed
  hp.gene.map <- results$cons.gene.map[results$cons.gene.map %in% genes]
  nhps <- length(hp.gene.map)
  hp.tab <- data.frame(shRNA=names(hp.gene.map), gene.symbol=hp.gene.map, isUsed=rep(FALSE, nhps),
                       sol.number=rep(NA, nhps), sol.name=rep(NA, nhps), sol.id=rep(NA, nhps),
                       row.names=names(hp.gene.map), check.rows=T, check.names=T, stringsAsFactors=F)

  hp.tab <- hp.tab[order(hp.tab[,"gene.symbol"], hp.tab[,"shRNA"]), ]
  
  for (g in genes) {
    g.solutions <- results$gene.solutions[[g]]
    if (g.solutions$num.solutions > 0) {
      for (s in g.solutions$solutions) {
        hps <- s$included.hps
        hp.tab[hps, "isUsed"] <- TRUE
        hp.tab[hps, "sol.number"] <- s$sol.num
        hp.tab[hps, "sol.name"] <- s$name
        hp.tab[hps, "sol.id"] <- s$id
      }
    }
  }

  hp.tab
}


############################################################################
# Read one or multple files and merge them in a single union-construct space  
RIGER.loadFiles <- function(params) {
with(params, {
  cat(paste("Loading file(s)",  shRNA.files, ":\n"))
  if ((file.format == "RES") || (file.format == "GCT")) {  

     if (file.format == "RES") {  # Read data from a single RES file
        dataset <- MSIG.Res2Frame(filename = shRNA.files)  # read only file
     } else {                     # Read data from a single GCT file
       if (length(grep('.Rdata$', shRNA.files)) > 0) {
         varname <- load(shRNA.files)
         stopifnot(varname == "gct") # this should be the name of the saved variable.
         dataset <- gct
         rm(gct)
       } else {  
         dataset <- MSIG.Gct2Frame(filename = shRNA.files)  # read only file
      }
#        dataset$ds[dataset$ds=="NA"] = NA # Aviad 09/06/05
     }
     constructs.set <- dataset$row.names
     sample.names <- dataset$names
     A <- dataset$ds
     descs <- dataset$descs
     if (sum(is.na(descs)) > 0) {
       cat(paste('***Warning:', sum(is.na(descs)), 'gene names where found to be NAs. Changing to empty string. ***\n'))
       descs[is.na(descs)] <- ""
     }
     dim(A) 
     cols <- length(A[1,])
     rows <- length(A[,1])
     gene.iden <- vector(length=rows, mode="character")
     for (i in 1:rows) {
        temp <- unlist(strsplit(noquote(descs[i]), ":"))
        if (length(temp) >= 1) {
          gene.iden[i] <- temp[[1]]
        }
     }
 
####

#     n.files <- length(shRNA.files)
#     constructs.set <- NULL
#     sample.names <- vector(length=n.files, mode = "character")
   cons.amplicon <- NULL

   for (i in exclude.cons) {
      locs <- seq(1:rows)[constructs.set == i]  # exclude constructs
      if (length(locs) >= 1) {
          A <- A[-locs,]
          gene.iden <- gene.iden[ -locs]
          constructs.set <- constructs.set[ - locs]
      }
      string <- paste("Number of excluded '", i, "' cons entries:", length(locs), '\n', sep='')   
      cat(string)
   }
   rows <- length(A[,1])      

   for (i in exclude.genes) {
      locs <- seq(1:rows)[gene.iden == i]  # exclude genes
      if (length(locs) >= 1) {
         A <- A[-locs,]
         gene.iden <- gene.iden[ -locs]
         constructs.set <- constructs.set[ - locs]
      }
      string <- paste("Number of excluded '", i, "' gene entries:", length(locs), '\n', sep='')
      cat(string)
   }
   rows <- length(A[,1])      

     hps.with.all.NAs <- which(apply(A, 1, function(x) all(is.na(x))))
     locs <- hps.with.all.NAs
     if (length(locs) >= 1) {
       A <- A[-locs,]
       gene.iden <- gene.iden[ -locs]
       constructs.set <- constructs.set[ - locs]
       string <- paste("Hairpins excluded can were all NAs:", length(locs), '\n', sep='')
       cat(string)
     }
     rows <- length(A[,1])      
   
   genes.with.only.one.cons <-  row.names(table(gene.iden))[table(gene.iden) == 1]
   if (length(genes.with.only.one.cons) >= 1) {
       locs <- match(genes.with.only.one.cons, gene.iden)
       A <- A[-locs,]
       gene.iden <- gene.iden[ -locs]
       constructs.set <- constructs.set[ - locs]
       string <- paste("Genes excluded because had only one construct:", length(locs), '\n', sep='')
       cat(string)
     }
   rows <- length(A[,1])      

     
   # Separate controls 
   # For now use all of the gene values instead of explicit controls

   # Rescale values per sample

   merged.array <- as.matrix(matrix(NA, nrow=rows, ncol=cols))
   num.cons <- length(constructs.set)
   cons.gene.map  <- vector(length=num.cons, mode="character")

   for (k in 1:cols) {
      if (norm.type == "standardize") {
         cntrl.median <- median(A[,k], na.rm=T)
         cntrl.mad <- mad(A[, k], na.rm=T)
         norm.vals <- (A[, k] - cntrl.median)/cntrl.mad
       } else if (norm.type == "shift") {
         norm.vals <- A[, k] - 1
       } else if (norm.type == "log") {
         cntrl.median <- median(A[, k])
         norm.vals <- log(A[, k]/cntrl.median)
       } else if (norm.type == "rescale") {
         cntrl.median <- median(A[, k])
         norm.vals <- (A[, k]/cntrl.median) - 1
       } else if (norm.type == "rescale2") {
         cntrl.median <- median(A[, k])
         norm.vals <- vector(length=rows, mode="numeric")
         for (h in 1:rows) {
             if (A[, k] >= cntrl.median) {
                norm.vals[h] <- A[h, k]/cntrl.median - 1
             } else {
                norm.vals[h] <- - (cntrl.median/A[h, k] + 1)
             }
         }
      } else {
         norm.vals <- A[, k]
      }
      norm.vals[norm.vals < -trunc.val] <- -trunc.val
      norm.vals[norm.vals >  trunc.val] <- trunc.val
      merged.array[, k] <- norm.vals
      cons.gene.map <- gene.iden
      
#     plot control and all construct histograms

      if (cntrl.plot == T) {
         den1 <- density(norm.vals, n = 512, from=min(norm.vals, na.rm=T), to=max(norm.vals, na.rm=T), adjust=1, na.rm = T)
         if (k == 1) {
           pdf(file = paste(results.dir, results.prefix, "_constructs_and_controls_dens_plot.pdf", sep=""), width = 13, height = 8)

#           X11(width = 14, height = 8)
            x.range <- c(min(den1$x, na.rm=T), max(den1$x, na.rm=T))
            y.range <- c(0, max(den1$y, na.rm=T))
            plot(den1$x, den1$y, type = "l", xlab = "Normalized Survival Score", ylab = "Prob.",
              main = "Prob density for constructs", lwd = 1, lty = "solid",
              col = 1, xlim = x.range, ylim = y.range)
              legend(x=max(den1$x, na.rm=T)*0.7, y=max(den1$y, na.rm=T)*0.7, legend=c("GENES"), 
                     bty="n", pch =22, bg = "white", pt.bg = 1, col = 1, cex = 1)
          } else { 
             points(den1$x, den1$y, type = "l", lwd = 1, lty = "solid", col = 1)
          }
      }
    } # end of loop over columns

### Multiple files

   } else if (file.format == "David" || file.format == "Milan") {

   # Compute size of array holding all files data excluding cntrl genes and gene with only one construct
  
   n.files <- length(shRNA.files)
   constructs.set <- NULL
   sample.names <- vector(length=n.files, mode = "character")

   k <- 1
   for (file in shRNA.files) {
      temp <- unlist(strsplit(x=file, split="/"))
      temp2 <- unlist(strsplit(x=temp[[length(temp)]], split="_"))
      sample.names[k] <- temp2[[1]]
     
      if (file.format == "Milan") {
        
         # clone name	symbol	NCBI gene id	MCR Amp #	Average luciferase units
         # NM_000026.1-460s1c1	ADSL	158	999	806047.5
         # NM_000026.1-1067s1c1	ADSL	158	999	1511382
         # NM_000026.1-873s1c1	ADSL	158	999	1497753.25
         # NM_000026.1-1042s1c1	ADSL	158	999	2088591
         # NM_020831.2-1735s1c1	MKL1	57591	999	1284875.25
         # ...

         mtab <- read.delim(file, header=T, sep="\t", blank.lines.skip=T, comment.char="", quote ="", as.is=T)
         mtab2 <- mtab[, c(1, 2, 5)]
         cons.amplicon <- mtab[, c(2, 4)]
         
      } else { # David's format
          
         # David's file format:

         #BRATS software version:	1	
         #Analysis date:	4/23/07	
         #Project:	Achilles-Renal	
         #Screen:	786 HK0106	
         #Batch:	786_ALL	
         #Batch date:	12/5/06	
         #Subset selection:	None (all batch plates)	
         #Scoring:	B-score	
         #Lower rejection threshold:	none	
         #Upper rejection threshold:	none	
         #Results grouped by:	HAIRPIN	

         # hairpin ID	gene Name	mean score
         # clonetechGfp_117s1c1	GFP	0.386176081
         # clonetechGfp_117s1c1	GFP	0.956763653
         # clonetechGfp_117s1c1	GFP	1.108812589
         # clonetechGfp_117s1c1	GFP	1.204238153
         # clonetechGfp_128s1c1	GFP	-0.270985954

         mtab2 <- read.delim(file, header=T, skip = 12, sep="\t", blank.lines.skip=T, comment.char="", quote ="", as.is=T)
         cons.amplicon <- NULL
      }

      genes.with.only.one.cons <- row.names(table(mtab2[,2]))[table(mtab2[,2]) == 1]
      if (length(genes.with.only.one.cons) >= 1) {
         locs <- match(genes.with.only.one.cons, mtab2[,2])
         mtab2 <- mtab2[-locs,]
         if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
          string <- paste("File:", file, " genes excluded because had only one construct:", length(locs), sep="")
         print(string)
      }
            
      for (i in exclude.cons) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,1] == i]  # exclude constructs
         if (length(locs) >= 1) {
             mtab2 <- mtab2[-locs,]
             if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
         }
         string <- paste("File:", file, " number of excluded ", i, " cons entries:", length(locs), sep=" ")
         print(string)
       }

      for (i in exclude.genes) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]  # exclude genes
         if (length(locs) >= 1) {
             mtab2 <- mtab2[-locs,]
             if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
         }
         string <- paste("File:", file, " number of excluded ", i, " gene entries:", length(locs), sep=" ")
         print(string)
       }

      for (i in cntrl.genes) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]
         if (length(locs) >= 1) {
             mtab2 <- mtab2[-locs,]
             if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
         }
         string <- paste("File:", file, " number of control ", i, " gene entries:", length(locs), sep=" ")
         print(string)
      }                      
      if (!is.null(cons.amplicon)) {
          locs2 <- seq(1:length(cons.amplicon[,2]))[cons.amplicon[,2] == 0] # amplicon = 0 are control genes too
          if (length(locs2) >= 1)  {
               mtab2 <- mtab2[-locs2,]
               cons.amplicon <- cons.amplicon[-locs2,]
          }
          string <- paste("File:", file, " number of amplicon 0 controls ", length(locs2), sep=" ")
          print(string)
      }

      mtab3 <- NULL
      if (!is.na(sub.select.genes)) {
         for (i in sub.select.genes) {
            locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]
            if (length(locs) >= 1) {
              if (is.null(mtab3)) {
                 mtab3 <- mtab2[locs,]
                 cons.amplicon2 <- cons.amplicon[locs,]
               } else {
                 mtab3 <- rbind(mtab3, mtab2[locs,])
                 cons.amplicon2 <- rbind(cons.amplicon2, cons.amplicon[locs,])
               }
            }
          }
       mtab2 <- mtab3
       cons.amplicon <- cons.amplicon2
      }
      print(c(k, file, "Number of constructs:", length(mtab2[, 1])))
      constructs.set <- c(constructs.set, mtab2[, 1])

      k <- k + 1
   }
   
   constructs.set <- unique(constructs.set)
   num.cons <- length(constructs.set)
   merged.array <- as.matrix(matrix(NA, nrow=num.cons, ncol=n.files))
   cons.gene.map  <- vector(length=num.cons, mode="character")
   
   k <- 1
   for (file in shRNA.files) {

      if (file.format == "Milan") {
         mtab <- read.delim(file, header=T, sep="\t", blank.lines.skip=T, comment.char="", quote ="", as.is=T)
         mtab2 <- mtab[, c(1, 2, 5)]
         cons.amplicon <- mtab[, c(2, 4)]

       } else { # David's format
#         mtab <- read.delim(file, header=T, sep="\t", blank.lines.skip=T, comment.char="", quote ="", as.is=T)
#         mtab2 <- mtab[-seq(1:11),]
         mtab2 <- read.delim(file, header=T, skip = 12, sep="\t", blank.lines.skip=T, comment.char="", quote ="", as.is=T)
         mtab2[, 3] <- as.numeric(noquote(mtab2[, 3]))
         cons.amplicon <- NULL
      }
      
      genes.with.only.one.cons <- row.names(table(mtab2[,2]))[table(mtab2[,2]) == 1]
      if (length(genes.with.only.one.cons) >= 1) {
         locs <- match(genes.with.only.one.cons, mtab2[,2])
        mtab2 <- mtab2[-locs,]
        if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
          string <- paste("File:", file, " genes excluded because had only one construct:", length(locs), sep="")
          print(string)
       }
      
      for (i in exclude.cons) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,1] == i]  # exclude constructs
         if (length(locs) >= 1) {
             mtab2 <- mtab2[-locs,]
             if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
         }
         string <- paste("File:", file, " number of excluded ", i, " cons entries:", length(locs), sep=" ")
         print(string)
       }

      for (i in exclude.genes) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]  # exclude genes
         if (length(locs) >= 1) {
            mtab2 <- mtab2[-locs,]
            if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
          }
         string <- paste("File:", file, " number of excluded ", i, " gene entries:", length(locs), sep=" ")
         print(string)
       }

      # separate controls and rescale values

      cntrl.vals <- NULL
      cntrl.vals2 <- NULL
      for (i in cntrl.genes) {
         locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]
         if (length(locs) >= 1) {
            cntrl.vals <- c(cntrl.vals, as.numeric(noquote(mtab2[locs, 3])))
            mtab2 <- mtab2[-locs,]
            if (!is.null(cons.amplicon)) cons.amplicon <- cons.amplicon[-locs,]
          }
         string <- paste("File:", file, " number of control ", i, " gene entries:", length(locs), sep=" ")
         print(string)
      }
      if(!is.na(cntrl.genes2)) {
         for (i in cntrl.genes2) {
            locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]
            if (length(locs) >= 1) {
               cntrl.vals2 <- c(cntrl.vals2, as.numeric(noquote(mtab2[locs, 3])))
             }
            string <- paste("File:", file, " number of (secondary) control ", i, " gene entries:", length(locs), sep=" ")
            print(string)
         }
      }
      if (!is.null(cons.amplicon)) {
          locs2 <- seq(1:length(cons.amplicon[,2]))[cons.amplicon[,2] == 0] # amplicon = 0 are control genes too
          if (length(locs2) >= 1) {
            cntrl.vals <- c(cntrl.vals, as.numeric(noquote(mtab2[locs2, 3])))
            mtab2 <- mtab2[-locs2,]
            cons.amplicon <- cons.amplicon[-locs2,]
          }
         string <- paste("File:", file, " number of amplicon 0 controls ", length(locs2), sep=" ")
         print(string)
      }


      mtab3 <- NULL
      if (!is.na(sub.select.genes)) {
         for (i in sub.select.genes) {
            locs <- seq(1:length(mtab2[,2]))[mtab2[,2] == i]
            if (length(locs) >= 1) {
              if (is.null(mtab3)) {
                 mtab3 <- mtab2[locs,]
                 cons.amplicon2 <- cons.amplicon[locs,]
               } else {
                 mtab3 <- rbind(mtab3, mtab2[locs,])
                 cons.amplicon2 <- rbind(cons.amplicon2, cons.amplicon[locs,])
               }
            }
          }
       mtab2 <- mtab3
       cons.amplicon <- cons.amplicon2
      }

      print(c(k, file, "Number of constructs:", length(mtab2[, 1])))

      
      if (norm.type == "standardize") {
         cntrl.median <- median(cntrl.vals)
         cntrl.mad <- mad(cntrl.vals)
         norm.vals <- (mtab2[, 3] - cntrl.median)/cntrl.mad
       } else if (norm.type == "shift") {
         norm.vals <- mtab2[, 3] - 1
       } else if (norm.type == "log") {
         cntrl.median <- median(cntrl.vals)
         norm.vals <- log(mtab2[, 3]/cntrl.median)
       } else if (norm.type == "rescale") {
         cntrl.median <- median(cntrl.vals)
         norm.vals <- (mtab2[, 3]/cntrl.median) - 1
       } else if (norm.type == "rescale2") {
         cntrl.median <- median(cntrl.vals)
         norm.vals <- vector(length=length(mtab2[, 3]), mode="numeric")
         for (h in 1:length(mtab2[, 3])) {
             if (mtab2[h, 3] >= cntrl.median) {
                norm.vals[h] <- mtab2[h, 3]/cntrl.median - 1
             } else {
                norm.vals[h] <- - (cntrl.median/mtab2[h, 3] + 1)
             }
         }
      } else {
         norm.vals <- mtab2[, 3]
      }

      norm.vals[norm.vals < -trunc.val] <- -trunc.val
      norm.vals[norm.vals >  trunc.val] <- trunc.val
      mtab2[, 3] <- norm.vals
      locs <- match(mtab2[, 1], constructs.set)
      print(c(k, file, "Number of cons to merged array:", length(locs)))
      merged.array[locs, k] <- mtab2[, 3]
      cons.gene.map[locs] <- mtab2[, 2]
      
# plot control and all construct histograms
     
      if (norm.type == "standardize") {
          cntrl.vals <- (cntrl.vals - cntrl.median)/cntrl.mad
       } else if (norm.type == "shift") {
          cntrl.vals <- cntrl.vals - 1
       } else if (norm.type == "log") {
          cntrl.vals <- log(cntrl.vals/cntrl.median)
       } else if (norm.type == "rescale") {
          cntrl.vals <- (cntrl.vals/cntrl.median) - 1
       } else if (norm.type == "rescale2") {
         for (h in 1:length(cntrl.vals)) {
             if (cntrl.vals[h] >= cntrl.median) {
                cntrl.vals[h] <- cntrl.vals[h]/cntrl.median - 1
             } else {
                cntrl.vals[h] <- - (cntrl.median/cntrl.vals[h] + 1)
             }
         }
       }


      if(!is.na(cntrl.genes2)) {
         if (norm.type == "standardize") {
             cntrl.vals2 <- (cntrl.vals2 - cntrl.median)/cntrl.mad
          } else if (norm.type == "shift") {
             cntrl.vals2 <- cntrl.vals2 - 1
          } else if (norm.type == "log") {
             cntrl.vals2 <- log(cntrl.vals2/cntrl.median)
          } else if (norm.type == "rescale") {
             cntrl.vals2 <- (cntrl.vals2/cntrl.median) - 1
          } else if (norm.type == "rescale2") {
            for (h in 1:length(cntrl.vals2)) {
               if (cntrl.vals2[h] >= cntrl.median) {
                  cntrl.vals2[h] <- cntrl.vals2[h]/cntrl.median - 1
               } else {
                  cntrl.vals2[h] <- - (cntrl.median/cntrl.vals2[h] + 1)
               }
            }
          }
      }

      if (cntrl.plot == T) {
         den1 <- density(cntrl.vals, n = 512, from=-3, to=3, adjust=1)
         den2 <- density(norm.vals, n = 512, from=-3, to=3, adjust=1)
         if(!is.na(cntrl.genes2)) den3 <- density(cntrl.vals2, n = 512, from=-3, to=3, adjust=1)

         if (k == 1) {
           pdf(file = paste(results.dir, results.prefix, "_constructs_and_controls_dens_plot.pdf", sep=""), width = 13, height = 8)
#            X11(width = 13, height = 8)
            x.range <- c(1.5*min(den1$x), 1.5*max(den1$x))
            y.range <- c(0, max(den1$y*1.5))
            plot(den1$x, den1$y, type = "l", xlab = "Normalized Survival Score", ylab = "Prob.",
              main = "Prob density for controls and constructs", lwd = 1, lty = "solid",
              col = 1, xlim = x.range, ylim = y.range)
            points(den2$x, den2$y, type = "l", lwd = 1, lty = "solid", col = 2)
            if(!is.na(cntrl.genes2)) {
               points(den3$x, den3$y, type = "l", lwd = 1, lty = "solid", col = 3)
               legend(x=max(den1$x)*0.7, y=max(den1$y)*0.7, legend=c("CNTRL", "GENES", "CNTRL2"), bty="n", pch = rep(22, 3), bg = "white", 
                      pt.bg = c(1, 2, 3), col = 1, cex = 1)
            } else {
               legend(x=max(den1$x)*0.7, y=max(den1$y)*0.7, legend=c("CNTRL", "GENES"), bty="n", pch = rep(22, 2), bg = "white", 
                      pt.bg = c(1, 2), col = 1, cex = 1)
            }
          } else { 
             points(den1$x, den1$y, type = "l", lwd = 1, lty = "solid", col = 1)
             points(den2$x, den2$y, type = "l", lwd = 1, lty = "solid", col = 2)
             if(!is.na(cntrl.genes2)) points(den3$x, den3$y, type = "l", lwd = 1, lty = "solid", col = 3)
          }
      }
      k <- k + 1
    } # end of loop over files
   print(c("total number of constructs", length(mtab2[, 1])))
   print(c("number of constructs per file:", apply(!is.na(merged.array), 2, sum)))

} # end of if file.format (one or multiple files)
    
  
   if (cntrl.plot == T) {
     dev.off()
#     savePlot(filename = paste(results.dir, results.prefix, "_constructs_and_controls_dens_plot.jpg", sep=""),
#         type ="jpeg", device = dev.cur())
    }

  return(list(num.cons=num.cons, merged.array=merged.array, sample.names=sample.names, constructs.set=constructs.set, cons.gene.map=cons.gene.map,
              cons.amplicon=cons.amplicon))
}) #with params
}       ##### end of file loading


MSIG.Gct2Frame <- function(filename = "NULL") { 
#
# Reads a gene expression dataset in GCT format and converts it into an R data frame
#
   ds <- read.delim(filename, header=T, sep="\t", skip=2, row.names=1, blank.lines.skip=T, comment.char="", as.is=T, na.strings = "NA")
   descs <- ds[,1]
   ds <- ds[-1]
   row.names <- row.names(ds)
   names <- names(ds)
   return(list(ds = ds, row.names = row.names, descs = descs, names = names))
}


write.gct <- function(gct.data.frame, descs = "", filename) 
{
    f <- file(filename, "w")
    cat("#1.2", "\n", file = f, append = TRUE, sep = "")
    cat(dim(gct.data.frame)[1], "\t", dim(gct.data.frame)[2], "\n", file = f, append = TRUE, sep = "")
    cat("Name", "\t", file = f, append = TRUE, sep = "")
    cat("Description", file = f, append = TRUE, sep = "")

    names <- names(gct.data.frame)
    cat("\t", names[1], file = f, append = TRUE, sep = "")

    if (length(names) > 1) {
       for (j in 2:length(names)) {
           cat("\t", names[j], file = f, append = TRUE, sep = "")
       }
     }
    cat("\n", file = f, append = TRUE, sep = "\t")

    oldWarn <- options(warn = -1)
    m <- matrix(nrow = dim(gct.data.frame)[1], ncol = dim(gct.data.frame)[2] +  2)
    m[, 1] <- row.names(gct.data.frame)
    if (length(descs) > 1) {
        m[, 2] <- descs
    } else {
        m[, 2] <- row.names(gct.data.frame)
    }
    index <- 3
    for (i in 1:dim(gct.data.frame)[2]) {
        m[, index] <- gct.data.frame[, i]
        index <- index + 1
    }
    write.table(m, file = f, append = TRUE, quote = FALSE, sep = "\t", eol = "\n", col.names = FALSE, row.names = FALSE)
    close(f)
    options(warn = 0)

}



# hp.data - hp data in each column, rownames - hp IDs.
# summarization.method - ("max" - max over hps of hp-hp corr, "combined.pval" - combine individual hp-hp 
# pvalues into one pvalue and use that as a score)
# corr.method  - correlation method to use (spearman, pearson, kendall)
# null.ecdf - an ecdf function for the null hp-hp corr scores. used by "spearman.combined.pval" method.
# solution.assignment - optional, vector of factors. specifies for each hairpin the solution it was used to
# create (all hairpins with no solution will have the same ID). If given, for each hairpin only the pvals with the
# hairpins assigned to the same solution are used to compute a combined p-value.
compCScore <- function(hp.data, summarization.method, corr.method,
                       null.pval.dist=NULL, solution.assignment=rep(1, nrow(hp.data))) {
  if (is.null(nrow(hp.data))) {
    NA
  } else {

    if (corr.method == "MI") { # Mutual Information similarity
      gcors <- mutual.inf.mat(t(hp.data), full=T)
    } else {
      gcors <- cor(t(hp.data), method=corr.method, use="pairwise.complete.obs")
    }
    
    if (summarization.method == "max") {
      diag(gcors) <- NA
      cscores <- apply(gcors, 1, max, na.rm=T)
      cor.pvals <- gcors
    } else if (summarization.method == "combined.pval") {
      require(survcomp, quietly=TRUE, warn.conflicts=FALSE)
      
      cors.pval <- matrix(1-null.pval.dist(gcors), nrow=nrow(gcors), ncol=ncol(gcors),
                          dimnames=dimnames(gcors))
      diag(cors.pval) <- NA
      # combine the pvalues for all hp-hp scores into one pvalues score using Stouffer's Z
      combine.method <- "z.transform"
      cscores <- apply(cors.pval, 2, combine.test, na.rm=T, method=combine.method)
      cor.pvals <- cors.pval
    }

    list(cscores=cscores, cor.pvals=cor.pvals)

  }
}


# This function will compute the Cscores for all hps of one gene, taking into account the grouping of hps by
# gene solutions; all hps of one solution are compared to each other, hps of no solution get a cscore by combining
# pvalues from all the gene hps.
compGeneCScoreBySolution <- function(hp.data, summarization.method="combined.pval", corr.method="MI",
                       null.pval.dist=NULL, solution.assignment=rep(1, nrow(hp.data))) {
  nhps <- nrow(hp.data)
  cscores <- vector(mode="numeric", length=nhps)
  hp.names <- rownames(hp.data)
  names(cscores) <- hp.names
  
  for (gs in unique(solution.assignment)) {
    hps <- hp.names[solution.assignment == gs]
    if (gs=="NA") {
      for (hp in hps) {
        rest.hps <- setdiff(hp.names, hp)
        if (corr.method == "MI") { # Mutual Information similarity
          gcors <- mutual.inf.v.mat(hp.data[hp,], t(hp.data[rest.hps,]))
        } else {
          gcors <- cor(matrix(hp.data[hp,], ncol=1),
                       matrix(t(hp.data[rest.hps,]), nrow=ncol(hp.data)),
                       method=corr.method, use="pairwise.complete.obs")
        }

        if (summarization.method == "max") {
          cscores[hp] <- max(gcors, na.rm=T)
        } else if (summarization.method == "combined.pval") {
          require(survcomp, quietly=TRUE, warn.conflicts=FALSE)
      
          cors.pval <- 1-null.pval.dist(gcors)
          # combine the pvalues for all hp-hp scores into one pvalues score using Stouffer's Z
          combine.method <- "z.transform"
          cscores[hp] <- combine.test(cors.pval, na.rm=T, method=combine.method)
        }
#        cat(paste(cscores[hp], "\n"))
      } # for hp in hps

    } else {
      cscores[hps] <- compCScore(hp.data[hps,], summarization.method, corr.method, null.pval.dist)$cscores
      
    }

  }

  cscores
}



# Given a matrix of values for each hairpin in a row and a vector of gene IDs with names shRNA IDs, compute
# consistensy scores (c-scores) for all shRNAs.
#
# hp.data - matrix, rownames=hp IDs, each row - data for one hp, columns - samples
# cons.gene.map - vector, values = gene IDs, names = hp IDs
#
compCScoreAllGenes <- function(hp.data, cons.gene.map, summarization.method,
                               corr.method, num.null.values=10000, num.cores=1, hp.table=NULL) {
  require(multicore, quietly=TRUE, warn.conflicts=FALSE)
  
  nh <- length(cons.gene.map)
  hpIDs <- names(cons.gene.map)
  geneIDs <- unique(cons.gene.map)
  null.cc <- NULL # used only for summarization.method="combined.pval"

  cscores <- matrix(nrow=nh, ncol=3, dimnames=list(hpIDs, c("cscore", "pval", "qval")))
  gene.cor.pvals <- list() # an item for each gene with correlation p-values for all hp pairs

  if (!is.null(hp.table)) {
    hp.table[is.na(hp.table[, "sol.number"]), "sol.number"] <- "NA"
  }

  # create null dist for hp-hp corr
  if (summarization.method == "combined.pval") {
    num.rand.hps <- ceiling((1+sqrt(1+8*num.null.values))/2) # solution of quadratic formula for (x^2 - x)/2=n
    num.rand.hps <- min(num.rand.hps, nrow(hp.data)) # we want to sample without replacement
    cat(paste("Computing all pairwise similarity scores for", num.rand.hps, "random hairpins to create null dist..."))

    if (corr.method == "MI") {
      cc <- mutual.inf.mat(t(hp.data[sample(nrow(hp.data), num.rand.hps, replace=F),]), full=F)
    } else {
      cc <- cor(t(hp.data[sample(nrow(hp.data), num.rand.hps, replace=F),]), method=corr.method, use="pairwise.complete.obs")
    }
    
    cc <- c(2, cc[upper.tri(cc)]) # add one sample of value > 1 so we'll never have ecdf(x) = 1 (so p-val won't be Inf)
    null.cc <- ecdf(cc)
    cat("Done.\n")
  }
  
  inside.func <- function(g) {
#    if (runif(1) < 0.01)
#      print(g)
    hps <- names(cons.gene.map[cons.gene.map == g])
    if (!is.null(hp.table)) {
      solution.assignment <- hp.table[hps, "sol.number"]
    }
    cscores <- compGeneCScoreBySolution(hp.data=hp.data[hps,],
                                        summarization.method=summarization.method,
                                        corr.method=corr.method,
                                        null.pval.dist=null.cc, solution.assignment=solution.assignment)
    list(hps=hps, cscores=cscores, g=g)
  } # inside.func

  r <- mclapply(geneIDs, inside.func, mc.cores=num.cores)

  for (gene in r) {
    cscores[gene$hps, "cscore"] <- signif(gene$cscores, 3)
    gene.cor.pvals[[gene$g]] <- gene$cor.pvals
  }

  
  # compute p/q-values
  gene.sizes <- sort(unique(table(cons.gene.map)))
  if (summarization.method == "max") {
    null.cscores <- list()
    require(multicore, quietly=TRUE, warn.conflicts=FALSE)
    null.cscores[gene.sizes] <- mclapply(gene.sizes,
                                         function(s) compCScoreRandomGenes(hp.data,
                                                                           nvalues=num.null.values,
                                                                           gene.size=s,
                                                                           corr.method=corr.method),
                                         mc.cores=num.cores)
  }

  
  for (s in gene.sizes) {
    sizes <- table(cons.gene.map)
    genes <- names(sizes[sizes==s])
    hps <- names(cons.gene.map[cons.gene.map %in% genes])

    if (summarization.method == "max") {
      cscores[hps, "pval"] <- as.numeric(sapply(hps, function(hp) { mean(cscores[hp, "cscore"] < null.cscores[[s]]) } ))
    } else if (summarization.method == "combined.pval") {
      cscores[hps, "pval"] <- cscores[hps, "cscore"]
      cscores[hps, "cscore"] <- signif(-log10(cscores[hps, "pval"]), 3)
    }
    
    cscores[hps, "qval"] <- signif(p.adjust(as.numeric(cscores[hps, "pval"]), method="BH"), 3)
  }
    
  cscores <- as.data.frame(cscores)
  cscores$symbol <- cons.gene.map[hpIDs]

  list(cscores=cscores, gene.cor.pvals=gene.cor.pvals)
}


# samples without replacement gene.size hps (rows) from hp.data, and compute CScore for each. repeats to create
# nsamples values.
compCScoreRandomGenes <- function(hp.data, nvalues=100, gene.size=5, corr.method="spearman") {
  cat(paste("Computing", nvalues, "null c-scores for genes of size", gene.size, "using", corr.method, "..."))
  nsamples <- ceiling(nvalues / gene.size)
  nh <- nrow(hp.data)
  null.cscore <- vector(mode="numeric", length=nsamples*gene.size)
  
  for (i in 1:nsamples) {
    hps <- sample(nh, gene.size, replace=F)
    null.cscore[((i-1) * gene.size + 1):(i*gene.size)] <- compCScore(hp.data[hps,],
                                                                     summarization.method="max",
                                                                     corr.method=corr.method)
  }
  
  cat("Done.\n")
  null.cscore
}


####### factorize.l1 #########
factorize.l1 <- function(m, num.iter = 20, epsilon = 0.005, max.gamma = 0.2, weighted=F, initA=NULL, initG=NULL,
                         positive.As=F) {
  require(quantreg, quietly=TRUE, warn.conflicts=FALSE)
  
  nh <- nrow(m)
  ns <- ncol(m)
  weights <- matrix(1/(nh*ns), nrow=nh, ncol=ns)
  A <- initA
  G <- initG

  if (is.null(initA)) {
    # randomly initialize initial values (for the Gs)
    # G <- rnorm(ns)
    G <- apply(m, 2, mean, na.rm=T)
    A <- rep(0, nh)
  }
  
  beta.h <- rep(0, nh)
  last.err <- Inf
  best <- list(objective=Inf)
 
  org.m <- m # save a copy of the input m matrix (with no imputed values)
  to.fit <- m # this is really what we're interested in fitting.
    
  # the actual model fitting iterations
  for (i in 1:num.iter) {

    sol <- factorize.l1.iteration(to.fit, A, G, weights, positive.As)
    A <- sol$A
    G <- sol$G

    # evaluate fitting
    GA <- outer(A, G, "*")
    pred.m <- GA  # our prediction for m

    resid <- to.fit - pred.m
    sigma <- mad(resid, na.rm=T)

    err <- sum(weights * abs(resid), na.rm=T)
#    cat(paste("Iteration", i,'- loss function is', err, '\n'))
#    cat(paste("Last err:",last.err,'\n'))

    if (err < best$objective) {
      best <- list(G=G, A=A, beta.h=beta.h, objective=err, i=i, m=org.m,
                   pred.m=pred.m, weights=weights, resid=resid)
    }    

    if (abs(1 - (err/last.err)) < epsilon || err < 0.001) {  # we have reached convergence
#      print("Acutally converged!")
      break
    }

#    stopifnot(err < last.err)
       
    last.err <- err

    if (weighted) {
      weights <- compute.weights(resid, sigma)
    }
  
  } # loop iterations

   
  if (i == num.iter)
    cat(paste("O-oh.", num.iter, "iterations weren't enough\n"))

  return(best)
}


# one iteration of weighted L1 bi-linear fitting of matrix 'to.fit'
factorize.l1.iteration <- function(to.fit, initA, initG, weights, positive.As=F) {
  require(aroma.light, quietly=TRUE, warn.conflicts=FALSE)
  nh <- nrow(to.fit)
  ns <- ncol(to.fit)
  A <- initA
  G <- initG

    for (h in 1:nh) {
      if (positive.As) {
#        A[h] <- rq(to.fit[h,] ~ G - 1, na.action='na.omit', tau=0.5, weights=weights[h,],
#                   method="fnc", R=matrix(1), r=0)$coefficients
#        A[h] <- weightedMedian(to.fit[h,] / G, w=abs(G) * weights[h,], ties="weighted", interpolate=F, na.rm=T)
        A[h] <- weightedMedian(to.fit[h,] / G, w=abs(G) * weights[h,], ties=NULL, interpolate=T, na.rm=T)
        A[h] <- max(A[h], 0, na.rm=T)
        
      } else {
#        A[h] <- rq(to.fit[h,] ~ G - 1, na.action='na.omit', tau=0.5, weights=weights[h,],
#                   method="br")$coefficients
#        A[h] <- weightedMedian(to.fit[h,] / G, w=abs(G) * weights[h,], ties="weighted", interpolate=F, na.rm=T)
        A[h] <- weightedMedian(to.fit[h,] / G, w=abs(G) * weights[h,], ties=NULL, interpolate=T, na.rm=T)
#        if ((A[h] - tmp) / A[h] > 0.01) {
#          print(A[h])
#          print(tmp)
#        }
      }
    }
    
    for (s in 1:ns) {
#      G[s] <- rq(to.fit[,s] ~ A - 1,  na.action='na.omit', tau=0.5, weights=weights[,s], method="br")$coefficients
#      G[s] <- weightedMedian(to.fit[,s] / A, w=abs(A) * weights[,s], ties="weighted", interpolate=F, na.rm=T)
      G[s] <- weightedMedian(to.fit[,s] / A, w=abs(A) * weights[,s], ties=NULL, interpolate=T, na.rm=T)

#      if (G[s] != 0 && (G[s] - tmp) / G[s] > 0.01) {
#        print(G[s])
#        print(tmp)
#      }

      
    }


  list(A=A, G=G)
}


# returns a weights matrix so that w_ij = c * exp(-(resid^2)/2*sigma^2) where resid is the element-wise residuals
# and sigma is the SD of the original data
compute.weights <- function(resid, sigma) {
  weights <- exp(-1 * (resid^2) / (2*sigma^2))

  weights <- weights / sum(weights)
  
  weights
}


# samples in columnd of 'to.fit', hps in rows.
wl1.fit.matrix <- function(to.fit, positive.As=F, weighted=T) {
  # compute robust location and scale
  #rscale <- apply(to.fit, 1, mad, na.rm=T)
  rscale <- rep(1, nrow(to.fit))
  names(rscale) <- rownames(to.fit)
  rmean <- apply(to.fit, 1, median, na.rm=T)

  # normalize input matrix
  #xs <- t(scale(t(to.fit), rmean, rscale))
  xs <- t(scale(t(to.fit), rmean, scale=F))
  
  f <- factorize.l1(xs, weighted=F, epsilon=0.01, positive.As=positive.As)

  if (weighted) {
    if (nrow(to.fit) > 2) { # weighted fitting is meaningless if only 2 hairpins are involved.
      f <- factorize.l1(xs, weighted=T, initA=f$A, initG=f$G, epsilon=0.005, positive.As=positive.As)
    }
  }   

  maxA <- max(abs(f$A), na.rm=T)
  f$A <- f$A / maxA
  f$G <- f$G * maxA

#  f$ressav <- xs - f$A %*% t(f$G)
  f$uniquenesses <- apply(f$resid, 1, mad, na.rm=T)
  f$MAD <- apply(to.fit, 1, mad, na.rm=T)
  f$xs <- xs
  f$scale <- rscale
  f$mean <- rmean

  f
}


l1.fit.matrix <- function(to.fit, positive.As=F) {
  wl1.fit.matrix(to.fit, positive.As, weighted=F)
}

#######
####### factorize.l1 #######

ATARiS(commandArgs(trailingOnly=T))

